Class: ViewComponent::Base
- Inherits:
-
Object
- Object
- ViewComponent::Base
- Includes:
- ActionView::Helpers, ActiveSupport::CoreExt::ERBUtil, ERB::Escape, InlineTemplate, Slotable, Translatable, WithContentHelper
- Defined in:
- lib/view_component/base.rb
Class Attribute Summary collapse
-
.identifier ⇒ String
The file path of the component Ruby file.
- .virtual_path ⇒ Object
Instance Attribute Summary collapse
-
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
-
#current_template ⇒ Object
readonly
Returns the value of attribute current_template.
Class Method Summary collapse
- .__vc_collection_counter_parameter ⇒ Object
- .__vc_collection_iteration_parameter ⇒ Object
- .__vc_collection_parameter ⇒ Object
- .__vc_compile(raise_errors: false, force: false) ⇒ Object
- .__vc_compiled? ⇒ Boolean
- .__vc_compiler ⇒ Object
- .__vc_counter_argument_present? ⇒ Boolean
- .__vc_ensure_compiled ⇒ Object
- .__vc_iteration_argument_present? ⇒ Boolean
-
.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter.
-
.__vc_validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn't define invalid parameters that could override the framework's methods.
-
.after_compile ⇒ Object
Hook called by the compiler after a component is compiled.
-
.config ⇒ ActiveSupport::OrderedOptions
Returns the current config.
- .inherited(child) ⇒ Object
-
.sidecar_files(extensions) ⇒ Object
Find sidecar files for the given extensions.
-
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
-
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
-
.with_collection(collection, spacer_component: nil, **args) ⇒ Object
Render a component for each element in a collection (documentation):.
-
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection (documentation):.
Instance Method Summary collapse
- #__vc_request ⇒ Object
-
#around_render ⇒ void
Called around rendering the component.
-
#before_render ⇒ void
Called before rendering the component.
-
#capture ⇒ Object
Sync @output_buffer with the view context's current output buffer before capturing.
-
#content ⇒ String
The content passed to the component instance as a block.
-
#content? ⇒ Boolean
Whether
contenthas been passed to the component. -
#controller ⇒ ActionController::Base
The current controller.
-
#format ⇒ Object
Rails expects us to define
formaton all renderables, but we do not know theformatof a ViewComponent until runtime. -
#helpers ⇒ ActionView::Base
A proxy through which to access helpers.
-
#initialize ⇒ Base
constructor
Including
Rails.application.routes.url_helpersdefines an initializer that accepts (...), so we have to define our own empty initializer to overwrite it. - #method_missing(method_name, *args) ⇒ Object
-
#output_postamble ⇒ String
Optional content to be returned after the rendered template.
-
#output_preamble ⇒ String
Optional content to be returned before the rendered template.
-
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we're not rendering a component.
-
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
-
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
-
#render_parent ⇒ Object
Subclass components that call
superinside their template code will cause a double render if they emit the result. -
#render_parent_to_string ⇒ Object
Renders the parent component to a string and returns it.
-
#request ⇒ ActionDispatch::Request
The current request.
-
#set_original_view_context(view_context) ⇒ void
Components render in their own view context.
-
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if.
-
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method.
-
#with_captured_virtual_path(captured_path) ⇒ Object
Temporarily sets the virtual path to the captured value, then restores it.
Methods included from WithContentHelper
Methods included from Translatable
Methods included from Slotable
#__vc_get_slot, #__vc_set_polymorphic_slot, #__vc_set_slot
Constructor Details
#initialize ⇒ Base
Including Rails.application.routes.url_helpers defines an initializer that accepts (...),
so we have to define our own empty initializer to overwrite it.
95 96 |
# File 'lib/view_component/base.rb', line 95 def initialize end |
Dynamic Method Handling
This class handles dynamic methods through the method_missing method
#method_missing(method_name, *args) ⇒ Object
342 343 344 345 346 347 348 349 350 351 352 353 |
# File 'lib/view_component/base.rb', line 342 def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing super rescue => e # rubocop:disable Style/RescueStandardError e.set_backtrace e.backtrace.tap(&:shift) raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name) #{e.} You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}`? MESSAGE raise end |
Class Attribute Details
.identifier ⇒ String
The file path of the component Ruby file.
573 574 575 |
# File 'lib/view_component/base.rb', line 573 def identifier @identifier end |
.virtual_path ⇒ Object
579 580 581 |
# File 'lib/view_component/base.rb', line 579 def virtual_path @virtual_path end |
Instance Attribute Details
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
74 75 76 |
# File 'lib/view_component/base.rb', line 74 def __vc_original_view_context @__vc_original_view_context end |
#current_template ⇒ Object (readonly)
Returns the value of attribute current_template.
75 76 77 |
# File 'lib/view_component/base.rb', line 75 def current_template @current_template end |
Class Method Details
.__vc_collection_counter_parameter ⇒ Object
773 774 775 |
# File 'lib/view_component/base.rb', line 773 def __vc_collection_counter_parameter @__vc_collection_counter_parameter ||= :"#{__vc_collection_parameter}_counter" end |
.__vc_collection_iteration_parameter ⇒ Object
783 784 785 |
# File 'lib/view_component/base.rb', line 783 def __vc_collection_iteration_parameter @__vc_collection_iteration_parameter ||= :"#{__vc_collection_parameter}_iteration" end |
.__vc_collection_parameter ⇒ Object
768 769 770 |
# File 'lib/view_component/base.rb', line 768 def __vc_collection_parameter @__vc_provided_collection_parameter ||= name && name.demodulize.underscore.chomp("_component").to_sym end |
.__vc_compile(raise_errors: false, force: false) ⇒ Object
638 639 640 |
# File 'lib/view_component/base.rb', line 638 def __vc_compile(raise_errors: false, force: false) __vc_compiler.compile(raise_errors: raise_errors, force: force) end |
.__vc_compiled? ⇒ Boolean
686 687 688 |
# File 'lib/view_component/base.rb', line 686 def __vc_compiled? __vc_compiler.compiled? end |
.__vc_compiler ⇒ Object
706 707 708 |
# File 'lib/view_component/base.rb', line 706 def __vc_compiler @__vc_compiler ||= Compiler.new(self) end |
.__vc_counter_argument_present? ⇒ Boolean
778 779 780 |
# File 'lib/view_component/base.rb', line 778 def __vc_counter_argument_present? __vc_initialize_parameter_names.include?(__vc_collection_counter_parameter) end |
.__vc_ensure_compiled ⇒ Object
701 702 703 |
# File 'lib/view_component/base.rb', line 701 def __vc_ensure_compiled __vc_compile unless __vc_compiled? end |
.__vc_iteration_argument_present? ⇒ Boolean
788 789 790 |
# File 'lib/view_component/base.rb', line 788 def __vc_iteration_argument_present? __vc_initialize_parameter_names.include?(__vc_collection_iteration_parameter) end |
.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter. By default, we don't validate that the default parameter name is accepted, as support for collection rendering is optional.
748 749 750 751 752 753 754 755 |
# File 'lib/view_component/base.rb', line 748 def __vc_validate_collection_parameter!(validate_default: false) parameter = validate_default ? __vc_collection_parameter : __vc_provided_collection_parameter return unless parameter return if __vc_initialize_parameter_names.include?(parameter) || __vc_splatted_keyword_argument_present? raise MissingCollectionArgumentError.new(name, parameter) end |
.__vc_validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn't define invalid parameters that could override the framework's methods.
761 762 763 764 765 |
# File 'lib/view_component/base.rb', line 761 def __vc_validate_initialization_parameters! return unless __vc_initialize_parameter_names.include?(:content) raise ReservedParameterError.new(name, :content) end |
.after_compile ⇒ Object
Hook called by the compiler after a component is compiled.
Extensions can override this class method to run logic after compilation (e.g., generate helpers, register metadata, etc.).
By default, this is a no-op.
696 697 698 |
# File 'lib/view_component/base.rb', line 696 def after_compile # no-op by default end |
.config ⇒ ActiveSupport::OrderedOptions
Returns the current config.
39 40 41 42 43 44 45 46 |
# File 'lib/view_component/base.rb', line 39 def config module_parents.each do |module_parent| next unless module_parent.respond_to?(:config) module_parent_config = module_parent.config.try(:view_component) return module_parent_config if module_parent_config end ViewComponent::Config.current end |
.inherited(child) ⇒ Object
643 644 645 646 647 648 649 650 651 652 653 654 655 656 657 658 659 660 661 662 663 664 665 666 667 668 669 670 671 672 673 674 675 676 677 678 679 680 681 682 683 |
# File 'lib/view_component/base.rb', line 643 def inherited(child) # Compile so child will inherit compiled `call_*` template methods that # `compile` defines __vc_compile # Give the child its own personal #render_template_for to protect against the case when # eager loading is disabled and the parent component is rendered before the child. In # such a scenario, the parent will override ViewComponent::Base#render_template_for, # meaning it will not be called for any children and thus not compile their templates. if !child.method_defined?(:render_template_for, false) && !child.__vc_compiled? child.class_eval <<~RUBY, __FILE__, __LINE__ + 1 def render_template_for(requested_details) # Force compilation here so the compiler always redefines render_template_for. # This is mostly a safeguard to prevent infinite recursion. self.class.__vc_compile(raise_errors: true, force: true) # .__vc_compile replaces this method; call the new one render_template_for(requested_details) end RUBY end # Derive the source location of the component Ruby file from the call stack. # We need to ignore `inherited` frames here as they indicate that `inherited` # has been re-defined by the consuming application, likely in ApplicationComponent. # We use `base_label` method here instead of `label` to avoid cases where the method # owner is included in a prefix like `ApplicationComponent.inherited`. child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path child.virtual_path = child.name&.underscore # Set collection parameter to the extended component child.with_collection_parameter(__vc_provided_collection_parameter) if method_defined?(:render_template_for, false) vc_ancestor_calls = defined?(@__vc_ancestor_calls) ? @__vc_ancestor_calls.dup : [] vc_ancestor_calls.unshift(instance_method(:render_template_for)) child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls) end super end |
.sidecar_files(extensions) ⇒ Object
Find sidecar files for the given extensions.
The provided array of extensions is expected to contain
strings starting without the dot, example: ["erb", "haml"].
For example, one might collect sidecar CSS files that need to be compiled.
588 589 590 591 592 593 594 595 596 597 598 599 600 601 602 603 604 605 606 607 608 609 610 611 612 613 614 615 616 617 618 619 620 621 622 |
# File 'lib/view_component/base.rb', line 588 def sidecar_files(extensions) return [] unless identifier return [] unless name extensions = extensions.join(",") # view files in a directory named like the component directory = File.dirname(identifier) filename = File.basename(identifier, ".rb") component_name = name.demodulize.underscore # Add support for nested components defined in the same file. # # for example # # class MyComponent < ViewComponent::Base # class MyOtherComponent < ViewComponent::Base # end # end # # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb` nested_component_files = if name.include?("::") && component_name != filename Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"] else [] end # view files in the same directory as the component sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"] sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"] (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq end |
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
class MyComponent < ViewComponent::Base
strip_trailing_whitespace
end
731 732 733 |
# File 'lib/view_component/base.rb', line 731 def strip_trailing_whitespace(value = true) self.__vc_strip_trailing_whitespace = value end |
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
738 739 740 |
# File 'lib/view_component/base.rb', line 738 def strip_trailing_whitespace? __vc_strip_trailing_whitespace end |
.with_collection(collection, spacer_component: nil, **args) ⇒ Object
Render a component for each element in a collection (documentation):
render(ProductsComponent.with_collection(@products, foo: :bar))
633 634 635 |
# File 'lib/view_component/base.rb', line 633 def with_collection(collection, spacer_component: nil, **args) Collection.new(self, collection, spacer_component, **args) end |
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection (documentation):
with_collection_parameter :item
717 718 719 720 |
# File 'lib/view_component/base.rb', line 717 def with_collection_parameter(parameter) @__vc_provided_collection_parameter = parameter @__vc_initialize_parameters = nil end |
Instance Method Details
#__vc_request ⇒ Object
386 387 388 389 |
# File 'lib/view_component/base.rb', line 386 def __vc_request # The current request (if present, as mailers/jobs/etc do not have a request) @__vc_request ||= controller.request if controller.respond_to?(:request) end |
#around_render ⇒ void
This method returns an undefined value.
Called around rendering the component. Override to wrap the rendering of a component in custom instrumentation, etc.
252 253 254 |
# File 'lib/view_component/base.rb', line 252 def around_render yield end |
#before_render ⇒ void
This method returns an undefined value.
Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.
244 245 246 |
# File 'lib/view_component/base.rb', line 244 def before_render # noop end |
#capture ⇒ Object
Sync @output_buffer with the view context's current output buffer before capturing. Form helpers create builders with @template_object pointing to the component. When those builders later call capture inside a partial (whose _run allocated a fresh OutputBuffer on the view context), the component's stale @output_buffer would capture from the wrong buffer. Temporarily switching to the view context's buffer keeps both in sync.
305 306 307 308 309 310 311 |
# File 'lib/view_component/base.rb', line 305 def capture(...) old_output_buffer = @output_buffer @output_buffer = view_context.output_buffer if view_context super ensure @output_buffer = old_output_buffer end |
#content ⇒ String
The content passed to the component instance as a block.
394 395 396 397 398 399 400 401 402 403 404 405 406 |
# File 'lib/view_component/base.rb', line 394 def content @__vc_content_evaluated = true return @__vc_content if defined?(@__vc_content) @__vc_content = if __vc_render_in_block_provided? with_captured_virtual_path(@old_virtual_path) do view_context.capture(self, &@__vc_render_in_block) end elsif __vc_content_set_by_with_content_defined? @__vc_content_set_by_with_content end end |
#content? ⇒ Boolean
Whether content has been passed to the component.
411 412 413 |
# File 'lib/view_component/base.rb', line 411 def content? __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined? end |
#controller ⇒ ActionController::Base
The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
317 318 319 320 321 |
# File 'lib/view_component/base.rb', line 317 def controller raise ControllerCalledBeforeRenderError if view_context.nil? @__vc_controller ||= view_context.controller end |
#format ⇒ Object
Rails expects us to define format on all renderables,
but we do not know the format of a ViewComponent until runtime.
372 373 374 |
# File 'lib/view_component/base.rb', line 372 def format nil end |
#helpers ⇒ ActionView::Base
A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
327 328 329 330 331 332 333 334 335 336 337 338 |
# File 'lib/view_component/base.rb', line 327 def helpers raise HelpersCalledBeforeRenderError if view_context.nil? # Attempt to re-use the original view_context passed to the first # component rendered in the rendering pipeline. This prevents the # instantiation of a new view_context via `controller.view_context` which # always returns a new instance of the view context class. # # This allows ivars to remain persisted when using the same helper via # `helpers` across multiple components and partials. @__vc_helpers ||= __vc_original_view_context || controller.view_context end |
#output_postamble ⇒ String
Optional content to be returned after the rendered template.
236 237 238 |
# File 'lib/view_component/base.rb', line 236 def output_postamble @@default_output_postamble ||= "".html_safe end |
#output_preamble ⇒ String
Optional content to be returned before the rendered template.
229 230 231 |
# File 'lib/view_component/base.rb', line 229 def output_preamble @@default_output_preamble ||= "".html_safe end |
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we're not rendering a component.
As of v4, ViewComponent::Base re-uses the existing view context created by ActionView, meaning the current view context and the original view context are the same object. set_original_view_context is still called to maintain backwards compatibility.
271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 |
# File 'lib/view_component/base.rb', line 271 def render( = {}, args = {}, &block) if .respond_to?(:set_original_view_context) .set_original_view_context(__vc_original_view_context) # We assume options is a component, so there's no need to evaluate the # block in the view context as we do below. @view_context.render(, args, &block) elsif block __vc_original_view_context.render(, args) do # capture the block output in the view context of the component output = capture(&block) # Partials are rendered to their own buffer and do not append to the # original @output_buffer we retain a reference to in #render_in. This # is a problem since the block passed to us here in the #render method # is evaluated within the context of ViewComponent::Base, and thus # appends to the original @output_buffer. To avoid this, we evaluate the # block in the view context instead, which will append to the output buffer # created for the partial. __vc_original_view_context.capture { output } end else __vc_original_view_context.render(, args) end end |
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
259 260 261 |
# File 'lib/view_component/base.rb', line 259 def render? true end |
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
view_context: ActionView context from calling viewblock: optional block to be captured within the view context
Returns HTML that has been escaped by the respective template handler.
106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 |
# File 'lib/view_component/base.rb', line 106 def render_in(view_context, **, &block) self.class.__vc_compile(raise_errors: true) unless self.class.__vc_compiled? __vc_check_reused_instance! __vc_reset_render_state! @view_context = view_context @old_virtual_path = view_context.instance_variable_get(:@virtual_path) self.__vc_original_view_context = view_context @output_buffer = view_context.output_buffer @lookup_context = view_context.lookup_context # For content_for @view_flow = view_context.view_flow # For i18n @virtual_path ||= virtual_path # Describes the inferred request constraints (locales, formats, variants) @__vc_requested_details = @lookup_context.vc_requested_details # For caching, such as #cache_if @current_template = nil unless defined?(@current_template) old_current_template = @current_template if block && defined?(@__vc_content_set_by_with_content) raise DuplicateContentError.new(self.class.name) end @__vc_content_evaluated = false remove_instance_variable(:@__vc_content) if defined?(@__vc_content) @__vc_render_in_block = block @view_context.instance_variable_set(:@virtual_path, virtual_path) before_render if render? value = nil @output_buffer.with_buffer do inner_rendered_template = nil around_rendered_template = around_render do inner_rendered_template = render_template_for(@__vc_requested_details).to_s end # If `around_render` returned the same object the block yielded, the inner # template's escaping is authoritative and we can trust the result. If the # user replaced/wrapped the value, re-check HTML safety to prevent # bypassing the escaping applied to normal `#call` return values # (GHSA-97jw-64cj-jc58). rendered_template = if around_rendered_template.equal?(inner_rendered_template) around_rendered_template else __vc_safe_around_render_output(around_rendered_template) end # Avoid allocating new string when output_preamble and output_postamble are blank value = if output_preamble.blank? && output_postamble.blank? rendered_template else __vc_safe_output_preamble + rendered_template + __vc_safe_output_postamble end end if ActionView::Base.annotate_rendered_view_with_filenames && current_template.inline_call? && request&.format == :html identifier = defined?(Rails.root) ? self.class.identifier.sub("#{Rails.root}/", "") : self.class.identifier value = "<!-- BEGIN #{identifier} -->".html_safe + value + "<!-- END #{identifier} -->".html_safe end value else "".html_safe end ensure view_context.instance_variable_set(:@virtual_path, @old_virtual_path) @current_template = old_current_template @__vc_rendered = true end |
#render_parent ⇒ Object
Subclass components that call super inside their template code will cause a
double render if they emit the result.
<%= super %> # double-renders
<% super %> # doesn't double-render
super also doesn't consider the current variant. render_parent renders the
parent template considering the current variant and emits the result without
double-rendering.
198 199 200 201 |
# File 'lib/view_component/base.rb', line 198 def render_parent render_parent_to_string nil end |
#render_parent_to_string ⇒ Object
Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.
def call
"<div>#{render_parent_to_string}</div>"
end
When rendering the parent inside an .erb template, use #render_parent instead.
213 214 215 216 217 218 219 220 221 222 223 224 |
# File 'lib/view_component/base.rb', line 213 def render_parent_to_string @__vc_parent_render_level ||= 0 # ensure a good starting value begin target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level] @__vc_parent_render_level += 1 target_render.bind_call(self, @__vc_requested_details) ensure @__vc_parent_render_level -= 1 end end |
#request ⇒ ActionDispatch::Request
The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
381 382 383 |
# File 'lib/view_component/base.rb', line 381 def request __vc_request end |
#set_original_view_context(view_context) ⇒ void
This method returns an undefined value.
Components render in their own view context. Helpers and other functionality
require a reference to the original Rails view context, an instance of
ActionView::Base. Use this method to set a reference to the original
view context. Objects that implement this method will render in the component's
view context, while objects that don't will render in the original view context
so helpers, etc work as expected.
87 88 89 |
# File 'lib/view_component/base.rb', line 87 def set_original_view_context(view_context) # noop end |
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if
365 366 367 |
# File 'lib/view_component/base.rb', line 365 def view_cache_dependencies [] end |
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method
359 360 361 |
# File 'lib/view_component/base.rb', line 359 def virtual_path self.class.virtual_path end |
#with_captured_virtual_path(captured_path) ⇒ Object
Temporarily sets the virtual path to the captured value, then restores it. This ensures translations and other path-dependent code execute with the correct scope.
418 419 420 421 422 423 424 |
# File 'lib/view_component/base.rb', line 418 def with_captured_virtual_path(captured_path) old_virtual_path = @view_context.instance_variable_get(:@virtual_path) @view_context.instance_variable_set(:@virtual_path, captured_path) yield ensure @view_context.instance_variable_set(:@virtual_path, old_virtual_path) end |