Class: ViewComponent::Base

Inherits:
Object
  • Object
show all
Includes:
ActionView::Helpers, ActiveSupport::CoreExt::ERBUtil, ERB::Escape, InlineTemplate, Slotable, Translatable, UseHelpers, WithContentHelper
Defined in:
lib/view_component/base.rb

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#__vc_i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Dynamic Method Handling

This class handles dynamic methods through the method_missing method

#method_missing(method_name, *args) ⇒ Object



274
275
276
277
278
279
280
281
282
283
284
285
# File 'lib/view_component/base.rb', line 274

def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing
  super
rescue => e # rubocop:disable Style/RescueStandardError
  e.set_backtrace e.backtrace.tap(&:shift)
  raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name)
    #{e.message}

    You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}`?
  MESSAGE

  raise
end

Class Attribute Details

.identifierString

The file path of the component Ruby file.

Returns:

  • (String)


451
452
453
# File 'lib/view_component/base.rb', line 451

def identifier
  @identifier
end

.virtual_pathObject



457
458
459
# File 'lib/view_component/base.rb', line 457

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



76
77
78
# File 'lib/view_component/base.rb', line 76

def __vc_original_view_context
  @__vc_original_view_context
end

#current_templateObject (readonly)

Returns the value of attribute current_template.



77
78
79
# File 'lib/view_component/base.rb', line 77

def current_template
  @current_template
end

Class Method Details

.__vc_collection_counter_parameterObject



653
654
655
# File 'lib/view_component/base.rb', line 653

def __vc_collection_counter_parameter
  :"#{__vc_collection_parameter}_counter"
end

.__vc_collection_iteration_parameterObject



663
664
665
# File 'lib/view_component/base.rb', line 663

def __vc_collection_iteration_parameter
  :"#{__vc_collection_parameter}_iteration"
end

.__vc_collection_parameterObject



648
649
650
# File 'lib/view_component/base.rb', line 648

def __vc_collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.__vc_compile(raise_errors: false, force: false) ⇒ Object



574
575
576
# File 'lib/view_component/base.rb', line 574

def __vc_compile(raise_errors: false, force: false)
  __vc_compiler.compile(raise_errors: raise_errors, force: force)
end

.__vc_compiled?Boolean

Returns:

  • (Boolean)


564
565
566
# File 'lib/view_component/base.rb', line 564

def __vc_compiled?
  __vc_compiler.compiled?
end

.__vc_compilerObject



579
580
581
# File 'lib/view_component/base.rb', line 579

def __vc_compiler
  @__vc_compiler ||= Compiler.new(self)
end

.__vc_counter_argument_present?Boolean

Returns:

  • (Boolean)


658
659
660
# File 'lib/view_component/base.rb', line 658

def __vc_counter_argument_present?
  initialize_parameter_names.include?(__vc_collection_counter_parameter)
end

.__vc_ensure_compiledObject



569
570
571
# File 'lib/view_component/base.rb', line 569

def __vc_ensure_compiled
  __vc_compile unless __vc_compiled?
end

.__vc_iteration_argument_present?Boolean

Returns:

  • (Boolean)


668
669
670
# File 'lib/view_component/base.rb', line 668

def __vc_iteration_argument_present?
  initialize_parameter_names.include?(__vc_collection_iteration_parameter)
end

.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
# File 'lib/view_component/base.rb', line 621

def __vc_validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? __vc_collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.__vc_validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



641
642
643
644
645
# File 'lib/view_component/base.rb', line 641

def __vc_validate_initialization_parameters!
  return unless initialize_parameter_names.include?(:content)

  raise ReservedParameterError.new(name, :content)
end

.configActiveSupport::OrderedOptions

Returns the current config.

Returns:

  • (ActiveSupport::OrderedOptions)


40
41
42
43
44
45
46
47
# File 'lib/view_component/base.rb', line 40

def config
  module_parents.each do |module_parent|
    next unless module_parent.respond_to?(:config)
    module_parent_config = module_parent.config.try(:view_component)
    return module_parent_config if module_parent_config
  end
  ViewComponent::Config.current
end

.inherited(child) ⇒ Object



515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
# File 'lib/view_component/base.rb', line 515

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  __vc_compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.__vc_compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(requested_details)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.__vc_compile(raise_errors: true, force: true)
        # .__vc_compile replaces this method; call the new one
        render_template_for(requested_details)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  # We use `base_label` method here instead of `label` to avoid cases where the method
  # owner is included in a prefix like `ApplicationComponent.inherited`.
  child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path
  child.virtual_path = child.name&.underscore

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  if instance_methods(false).include?(:render_template_for)
    vc_ancestor_calls = defined?(@__vc_ancestor_calls) ? @__vc_ancestor_calls.dup : []

    vc_ancestor_calls.unshift(instance_method(:render_template_for))
    child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls)
  end

  super
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
# File 'lib/view_component/base.rb', line 466

def sidecar_files(extensions)
  return [] unless identifier

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(identifier)
  filename = File.basename(identifier, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



604
605
606
# File 'lib/view_component/base.rb', line 604

def strip_trailing_whitespace(value = true)
  self.__vc_strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


611
612
613
# File 'lib/view_component/base.rb', line 611

def strip_trailing_whitespace?
  __vc_strip_trailing_whitespace
end

.with_collection(collection, spacer_component: nil, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • spacer_component (ViewComponent::Base) (defaults to: nil)

    Component instance to be rendered between items.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



510
511
512
# File 'lib/view_component/base.rb', line 510

def with_collection(collection, spacer_component: nil, **args)
  Collection.new(self, collection, spacer_component, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



590
591
592
593
# File 'lib/view_component/base.rb', line 590

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
  @initialize_parameters = nil
end

Instance Method Details

#__vc_requestObject



310
311
312
313
# File 'lib/view_component/base.rb', line 310

def __vc_request
  # The current request (if present, as mailers/jobs/etc do not have a request)
  @__vc_request ||= controller.request if controller.respond_to?(:request)
end

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



218
219
220
# File 'lib/view_component/base.rb', line 218

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


318
319
320
321
322
323
324
325
326
327
328
# File 'lib/view_component/base.rb', line 318

def content
  @__vc_content_evaluated = true
  return @__vc_content if defined?(@__vc_content)

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content_defined?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


333
334
335
# File 'lib/view_component/base.rb', line 333

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



249
250
251
252
253
# File 'lib/view_component/base.rb', line 249

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



259
260
261
262
263
264
265
266
267
268
269
270
# File 'lib/view_component/base.rb', line 259

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


210
211
212
# File 'lib/view_component/base.rb', line 210

def output_postamble
  @@default_output_postamble ||= "".html_safe
end

#output_preambleString

Optional content to be returned before the rendered template.

Returns:

  • (String)


203
204
205
# File 'lib/view_component/base.rb', line 203

def output_preamble
  @@default_output_preamble ||= "".html_safe
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`, and not the component’s ‘view_context`.



236
237
238
239
240
241
242
243
# File 'lib/view_component/base.rb', line 236

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    @view_context.render(options, args, &block)
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


225
226
227
# File 'lib/view_component/base.rb', line 225

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
# File 'lib/view_component/base.rb', line 102

def render_in(view_context, &block)
  self.class.__vc_compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = view_context.output_buffer

  @lookup_context ||= view_context.lookup_context

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # Describes the inferred request constraints (locales, formats, variants)
  @__vc_requested_details ||= @lookup_context.vc_requested_details

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template

  if block && defined?(@__vc_content_set_by_with_content)
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    value = nil

    @output_buffer.with_buffer do
      rendered_template = render_template_for(@__vc_requested_details).to_s

      # Avoid allocating new string when output_preamble and output_postamble are blank
      value = if output_preamble.blank? && output_postamble.blank?
        rendered_template
      else
        safe_output_preamble + rendered_template + safe_output_postamble
      end
    end

    if ActionView::Base.annotate_rendered_view_with_filenames && current_template.inline_call? && request&.format == :html
      identifier = defined?(Rails.root) ? self.class.identifier.sub("#{Rails.root}/", "") : self.class.identifier
      value = "<!-- BEGIN #{identifier} -->".html_safe + value + "<!-- END #{identifier} -->".html_safe
    end

    value
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.

“‘erb <%= super %> # double-renders <% super %> # doesn’t double-render “‘

‘super` also doesn’t consider the current variant. ‘render_parent` renders the parent template considering the current variant and emits the result without double-rendering.



172
173
174
175
# File 'lib/view_component/base.rb', line 172

def render_parent
  render_parent_to_string
  nil
end

#render_parent_to_stringObject

Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.

“‘ruby def call

"<div>#{render_parent_to_string}</div>"

end “‘

When rendering the parent inside an .erb template, use ‘#render_parent` instead.



187
188
189
190
191
192
193
194
195
196
197
198
# File 'lib/view_component/base.rb', line 187

def render_parent_to_string
  @__vc_parent_render_level ||= 0 # ensure a good starting value

  begin
    target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level]
    @__vc_parent_render_level += 1

    target_render.bind_call(self, @__vc_requested_details)
  ensure
    @__vc_parent_render_level -= 1
  end
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


305
306
307
# File 'lib/view_component/base.rb', line 305

def request
  __vc_request
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



88
89
90
# File 'lib/view_component/base.rb', line 88

def set_original_view_context(view_context)
  # noop
end

#view_cache_dependenciesObject

For caching, such as #cache_if



297
298
299
# File 'lib/view_component/base.rb', line 297

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



291
292
293
# File 'lib/view_component/base.rb', line 291

def virtual_path
  self.class.virtual_path
end