Class: ViewComponent::Base

Inherits:
Object
  • Object
show all
Includes:
ActionView::Helpers, ActiveSupport::CoreExt::ERBUtil, ERB::Escape, ComponentLocalConfig, InlineTemplate, Slotable, Translatable, UseHelpers, WithContentHelper
Defined in:
lib/view_component/base.rb

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#__vc_i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Dynamic Method Handling

This class handles dynamic methods through the method_missing method

#method_missing(method_name, *args) ⇒ Object



290
291
292
293
294
295
296
297
298
299
300
301
# File 'lib/view_component/base.rb', line 290

def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing
  super
rescue => e # rubocop:disable Style/RescueStandardError
  e.set_backtrace e.backtrace.tap(&:shift)
  raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name)
    #{e.message}

    You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}`?
  MESSAGE

  raise
end

Class Attribute Details

.identifierString

The file path of the component Ruby file.

Returns:

  • (String)


475
476
477
# File 'lib/view_component/base.rb', line 475

def identifier
  @identifier
end

.virtual_pathObject



481
482
483
# File 'lib/view_component/base.rb', line 481

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



103
104
105
# File 'lib/view_component/base.rb', line 103

def __vc_original_view_context
  @__vc_original_view_context
end

#current_templateObject (readonly)

Returns the value of attribute current_template.



104
105
106
# File 'lib/view_component/base.rb', line 104

def current_template
  @current_template
end

Class Method Details

.__vc_collection_counter_parameterObject



699
700
701
# File 'lib/view_component/base.rb', line 699

def __vc_collection_counter_parameter
  :"#{__vc_collection_parameter}_counter"
end

.__vc_collection_iteration_parameterObject



709
710
711
# File 'lib/view_component/base.rb', line 709

def __vc_collection_iteration_parameter
  :"#{__vc_collection_parameter}_iteration"
end

.__vc_collection_parameterObject



694
695
696
# File 'lib/view_component/base.rb', line 694

def __vc_collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.__vc_compile(raise_errors: false, force: false) ⇒ Object



598
599
600
# File 'lib/view_component/base.rb', line 598

def __vc_compile(raise_errors: false, force: false)
  __vc_compiler.compile(raise_errors: raise_errors, force: force)
end

.__vc_compiled?Boolean

Returns:

  • (Boolean)


588
589
590
# File 'lib/view_component/base.rb', line 588

def __vc_compiled?
  __vc_compiler.compiled?
end

.__vc_compilerObject



603
604
605
# File 'lib/view_component/base.rb', line 603

def __vc_compiler
  @__vc_compiler ||= Compiler.new(self)
end

.__vc_counter_argument_present?Boolean

Returns:

  • (Boolean)


704
705
706
# File 'lib/view_component/base.rb', line 704

def __vc_counter_argument_present?
  initialize_parameter_names.include?(__vc_collection_counter_parameter)
end

.__vc_ensure_compiledObject



593
594
595
# File 'lib/view_component/base.rb', line 593

def __vc_ensure_compiled
  __vc_compile unless __vc_compiled?
end

.__vc_iteration_argument_present?Boolean

Returns:

  • (Boolean)


714
715
716
# File 'lib/view_component/base.rb', line 714

def __vc_iteration_argument_present?
  initialize_parameter_names.include?(__vc_collection_iteration_parameter)
end

.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



667
668
669
670
671
672
673
674
675
676
677
678
679
680
681
# File 'lib/view_component/base.rb', line 667

def __vc_validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? __vc_collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.__vc_validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



687
688
689
690
691
# File 'lib/view_component/base.rb', line 687

def __vc_validate_initialization_parameters!
  return unless initialize_parameter_names.include?(:content)

  raise ReservedParameterError.new(name, :content)
end

.configActiveSupport::OrderedOptions

Returns the current config.

Returns:

  • (ActiveSupport::OrderedOptions)


50
51
52
53
54
55
56
57
# File 'lib/view_component/base.rb', line 50

def config
  module_parents.each do |module_parent|
    next unless module_parent.respond_to?(:config)
    module_parent_config = module_parent.config.try(:view_component)
    return module_parent_config if module_parent_config
  end
  ViewComponent::Config.current
end

.inherited(child) ⇒ Object



539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
# File 'lib/view_component/base.rb', line 539

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  __vc_compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.__vc_compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(requested_details)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.__vc_compile(raise_errors: true, force: true)
        # .__vc_compile replaces this method; call the new one
        render_template_for(requested_details)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  # We use `base_label` method here instead of `label` to avoid cases where the method
  # owner is included in a prefix like `ApplicationComponent.inherited`.
  child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path
  child.virtual_path = child.name&.underscore

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  if instance_methods(false).include?(:render_template_for)
    vc_ancestor_calls = (!@__vc_ancestor_calls.nil?) ? @__vc_ancestor_calls.dup : []

    vc_ancestor_calls.unshift(instance_method(:render_template_for))
    child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls)
  end

  super
end

.newObject

Redefine ‘new` so we can pre-allocate instance variables to optimize for Ruby object shapes.



40
41
42
43
44
45
# File 'lib/view_component/base.rb', line 40

def new(...)
  instance = allocate
  instance.__vc_pre_allocate_instance_variables
  instance.send(:initialize, ...)
  instance
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
# File 'lib/view_component/base.rb', line 490

def sidecar_files(extensions)
  return [] unless identifier

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(identifier)
  filename = File.basename(identifier, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Deprecated.

Use the new component-local configuration option instead.

“‘ruby class MyComponent < ViewComponent::Base

configure_view_component do |config|
  config.strip_trailing_whitespace = true
end

end “‘

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



638
639
640
641
642
643
644
645
646
647
648
649
650
651
652
# File 'lib/view_component/base.rb', line 638

def strip_trailing_whitespace(value = true)
  ViewComponent::Deprecation.deprecation_warning(
    "strip_trailing_whitespace",
    <<~DOC
      Use the new component-local configuration option instead:

      class #{self.class.name} < ViewComponent::Base
        configure_view_component do |config|
          config.strip_trailing_whitespace = #{value}
        end
      end
    DOC
  )
  view_component_config.strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


657
658
659
# File 'lib/view_component/base.rb', line 657

def strip_trailing_whitespace?
  view_component_config.strip_trailing_whitespace
end

.with_collection(collection, spacer_component: nil, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • spacer_component (ViewComponent::Base) (defaults to: nil)

    Component instance to be rendered between items.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



534
535
536
# File 'lib/view_component/base.rb', line 534

def with_collection(collection, spacer_component: nil, **args)
  Collection.new(self, collection, spacer_component, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



614
615
616
617
# File 'lib/view_component/base.rb', line 614

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
  @initialize_parameters = nil
end

Instance Method Details

#__vc_pre_allocate_instance_variablesObject



60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
# File 'lib/view_component/base.rb', line 60

def __vc_pre_allocate_instance_variables
  @__vc_parent_render_level = 0
  @__vc_set_slots = {}
  @__vc_content_evaluated = false
  @current_template = nil
  @output_buffer = nil
  @lookup_context = nil
  @view_flow = nil
  @view_context = nil
  @virtual_path = nil
  @__vc_ancestor_calls = nil
  @__vc_controller = nil
  @__vc_content = :unset # some behaviors depend on checking for nil
  @__vc_content_set_by_with_content = nil
  @__vc_helpers = nil
  @__vc_inline_template = nil
  @__vc_inline_template_defined = nil
  @__vc_render_in_block = nil
  @__vc_request = nil
  @__vc_requested_details = nil
  @__vc_original_view_context = nil
end

#__vc_requestObject



326
327
328
329
# File 'lib/view_component/base.rb', line 326

def __vc_request
  # The current request (if present, as mailers/jobs/etc do not have a request)
  @__vc_request ||= controller.request if controller.respond_to?(:request)
end

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



234
235
236
# File 'lib/view_component/base.rb', line 234

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


334
335
336
337
338
339
340
341
342
343
344
# File 'lib/view_component/base.rb', line 334

def content
  @__vc_content_evaluated = true
  return @__vc_content if @__vc_content != :unset

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


349
350
351
# File 'lib/view_component/base.rb', line 349

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



265
266
267
268
269
# File 'lib/view_component/base.rb', line 265

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



275
276
277
278
279
280
281
282
283
284
285
286
# File 'lib/view_component/base.rb', line 275

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


226
227
228
# File 'lib/view_component/base.rb', line 226

def output_postamble
  @@default_output_postamble ||= "".html_safe
end

#output_preambleString

Optional content to be returned before the rendered template.

Returns:

  • (String)


219
220
221
# File 'lib/view_component/base.rb', line 219

def output_preamble
  @@default_output_preamble ||= "".html_safe
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`, and not the component’s ‘view_context`.



252
253
254
255
256
257
258
259
# File 'lib/view_component/base.rb', line 252

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    @view_context.render(options, args, &block)
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


241
242
243
# File 'lib/view_component/base.rb', line 241

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
# File 'lib/view_component/base.rb', line 129

def render_in(view_context, &block)
  self.class.__vc_compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = view_context.output_buffer

  @lookup_context ||= view_context.lookup_context

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # Describes the inferred request constraints (locales, formats, variants)
  @__vc_requested_details ||= @lookup_context.vc_requested_details

  # For caching, such as #cache_if
  old_current_template = @current_template

  if block && __vc_content_set_by_with_content?
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_render_in_block = block

  before_render

  if render?
    value = nil

    @output_buffer.with_buffer do
      rendered_template = render_template_for(@__vc_requested_details).to_s

      # Avoid allocating new string when output_preamble and output_postamble are blank
      value = if output_preamble.blank? && output_postamble.blank?
        rendered_template
      else
        safe_output_preamble + rendered_template + safe_output_postamble
      end
    end

    value
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.

“‘erb <%= super %> # double-renders <% super %> # doesn’t double-render “‘

‘super` also doesn’t consider the current variant. ‘render_parent` renders the parent template considering the current variant and emits the result without double-rendering.



192
193
194
195
# File 'lib/view_component/base.rb', line 192

def render_parent
  render_parent_to_string
  nil
end

#render_parent_to_stringObject

Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.

“‘ruby def call

"<div>#{render_parent_to_string}</div>"

end “‘

When rendering the parent inside an .erb template, use ‘#render_parent` instead.



207
208
209
210
211
212
213
214
# File 'lib/view_component/base.rb', line 207

def render_parent_to_string
  target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level]
  @__vc_parent_render_level += 1

  target_render.bind_call(self, @__vc_requested_details)
ensure
  @__vc_parent_render_level -= 1
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


321
322
323
# File 'lib/view_component/base.rb', line 321

def request
  __vc_request
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



115
116
117
# File 'lib/view_component/base.rb', line 115

def set_original_view_context(view_context)
  # noop
end

#view_cache_dependenciesObject

For caching, such as #cache_if



313
314
315
# File 'lib/view_component/base.rb', line 313

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



307
308
309
# File 'lib/view_component/base.rb', line 307

def virtual_path
  self.class.virtual_path
end