Class: ViewComponent::Base

Inherits:
ActionView::Base
  • Object
show all
Includes:
InlineTemplate, Slotable, Translatable, UseHelpers, WithContentHelper
Defined in:
lib/view_component/base.rb

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#__vc_i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Constructor Details

#initializeBase

Override the ActionView::Base initializer so that components do not need to define their own initializers.



196
197
# File 'lib/view_component/base.rb', line 196

def initialize(*)
end

Dynamic Method Handling

This class handles dynamic methods through the method_missing method

#method_missing(method_name, *args) ⇒ Object



244
245
246
247
248
249
250
251
252
253
254
255
# File 'lib/view_component/base.rb', line 244

def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing
  super
rescue => e # rubocop:disable Style/RescueStandardError
  e.set_backtrace e.backtrace.tap(&:shift)
  raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name)
    #{e.message}

    You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}'?
  MESSAGE

  raise
end

Class Attribute Details

.identifierString

The file path of the component Ruby file.

Returns:

  • (String)


438
439
440
# File 'lib/view_component/base.rb', line 438

def identifier
  @identifier
end

.virtual_pathObject



444
445
446
# File 'lib/view_component/base.rb', line 444

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



52
53
54
# File 'lib/view_component/base.rb', line 52

def __vc_original_view_context
  @__vc_original_view_context
end

Class Method Details

.__vc_collection_counter_parameterObject



646
647
648
# File 'lib/view_component/base.rb', line 646

def __vc_collection_counter_parameter
  :"#{__vc_collection_parameter}_counter"
end

.__vc_collection_iteration_parameterObject



656
657
658
# File 'lib/view_component/base.rb', line 656

def __vc_collection_iteration_parameter
  :"#{__vc_collection_parameter}_iteration"
end

.__vc_collection_parameterObject



641
642
643
# File 'lib/view_component/base.rb', line 641

def __vc_collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.__vc_compile(raise_errors: false, force: false) ⇒ Object



567
568
569
# File 'lib/view_component/base.rb', line 567

def __vc_compile(raise_errors: false, force: false)
  __vc_compiler.compile(raise_errors: raise_errors, force: force)
end

.__vc_compiled?Boolean

Returns:

  • (Boolean)


557
558
559
# File 'lib/view_component/base.rb', line 557

def __vc_compiled?
  __vc_compiler.compiled?
end

.__vc_compilerObject



572
573
574
# File 'lib/view_component/base.rb', line 572

def __vc_compiler
  @__vc_compiler ||= Compiler.new(self)
end

.__vc_counter_argument_present?Boolean

Returns:

  • (Boolean)


651
652
653
# File 'lib/view_component/base.rb', line 651

def __vc_counter_argument_present?
  initialize_parameter_names.include?(__vc_collection_counter_parameter)
end

.__vc_ensure_compiledObject



562
563
564
# File 'lib/view_component/base.rb', line 562

def __vc_ensure_compiled
  __vc_compile unless __vc_compiled?
end

.__vc_iteration_argument_present?Boolean

Returns:

  • (Boolean)


661
662
663
# File 'lib/view_component/base.rb', line 661

def __vc_iteration_argument_present?
  initialize_parameter_names.include?(__vc_collection_iteration_parameter)
end

.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



614
615
616
617
618
619
620
621
622
623
624
625
626
627
628
# File 'lib/view_component/base.rb', line 614

def __vc_validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? __vc_collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.__vc_validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



634
635
636
637
638
# File 'lib/view_component/base.rb', line 634

def __vc_validate_initialization_parameters!
  return unless initialize_parameter_names.include?(:content)

  raise ReservedParameterError.new(name, :content)
end

.configActiveSupport::OrderedOptions

Returns the current config.

Returns:

  • (ActiveSupport::OrderedOptions)


28
29
30
31
32
33
34
# File 'lib/view_component/base.rb', line 28

def config
  module_parents.each do |m|
    config = m.try(:config).try(:view_component)
    return config if config
  end
  ViewComponent::Config.current
end

.inherited(child) ⇒ Object



502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
# File 'lib/view_component/base.rb', line 502

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  __vc_compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.__vc_compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(requested_details)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.__vc_compile(raise_errors: true, force: true)
        # .__vc_compile replaces this method; call the new one
        render_template_for(requested_details)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  # We use `base_label` method here instead of `label` to avoid cases where the method
  # owner is included in a prefix like `ApplicationComponent.inherited`.
  child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path

  # If Rails application is loaded, removes the first part of the path and the extension.
  if defined?(Rails) && Rails.application
    child.virtual_path = child.identifier.gsub(
      /(.*#{Regexp.quote(ViewComponent::Base.config.view_component_path)})|(\.rb)/, ""
    )
  end

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  if instance_methods(false).include?(:render_template_for)
    vc_ancestor_calls = defined?(@__vc_ancestor_calls) ? @__vc_ancestor_calls.dup : []

    vc_ancestor_calls.unshift(instance_method(:render_template_for))
    child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls)
  end

  super
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
# File 'lib/view_component/base.rb', line 453

def sidecar_files(extensions)
  return [] unless identifier

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(identifier)
  filename = File.basename(identifier, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



597
598
599
# File 'lib/view_component/base.rb', line 597

def strip_trailing_whitespace(value = true)
  self.__vc_strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


604
605
606
# File 'lib/view_component/base.rb', line 604

def strip_trailing_whitespace?
  __vc_strip_trailing_whitespace
end

.with_collection(collection, spacer_component: nil, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • spacer_component (ViewComponent::Base) (defaults to: nil)

    Component instance to be rendered between items.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



497
498
499
# File 'lib/view_component/base.rb', line 497

def with_collection(collection, spacer_component: nil, **args)
  Collection.new(self, collection, spacer_component, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



583
584
585
586
# File 'lib/view_component/base.rb', line 583

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
  @initialize_parameters = nil
end

Instance Method Details

#__vc_requestObject



280
281
282
283
# File 'lib/view_component/base.rb', line 280

def __vc_request
  # The current request (if present, as mailers/jobs/etc do not have a request)
  @__vc_request ||= controller.request if controller.respond_to?(:request)
end

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



182
183
184
# File 'lib/view_component/base.rb', line 182

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


288
289
290
291
292
293
294
295
296
297
298
# File 'lib/view_component/base.rb', line 288

def content
  @__vc_content_evaluated = true
  return @__vc_content if defined?(@__vc_content)

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content_defined?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


303
304
305
# File 'lib/view_component/base.rb', line 303

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



219
220
221
222
223
# File 'lib/view_component/base.rb', line 219

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



229
230
231
232
233
234
235
236
237
238
239
240
# File 'lib/view_component/base.rb', line 229

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


174
175
176
# File 'lib/view_component/base.rb', line 174

def output_postamble
  @@default_output_postamble ||= "".html_safe
end

#output_preambleString

Optional content to be returned before the rendered template.

Returns:

  • (String)


167
168
169
# File 'lib/view_component/base.rb', line 167

def output_preamble
  @@default_output_preamble ||= "".html_safe
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`, and not the component’s ‘view_context`.



206
207
208
209
210
211
212
213
# File 'lib/view_component/base.rb', line 206

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    super
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


189
190
191
# File 'lib/view_component/base.rb', line 189

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
# File 'lib/view_component/base.rb', line 77

def render_in(view_context, &block)
  self.class.__vc_compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = ActionView::OutputBuffer.new

  @lookup_context ||= view_context.lookup_context

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # Describes the inferred request constraints (locales, formats, variants)
  @__vc_requested_details ||= @lookup_context.vc_requested_details

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template

  if block && defined?(@__vc_content_set_by_with_content)
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    rendered_template = render_template_for(@__vc_requested_details).to_s

    # Avoid allocating new string when output_preamble and output_postamble are blank
    if output_preamble.blank? && output_postamble.blank?
      rendered_template
    else
      safe_output_preamble + rendered_template + safe_output_postamble
    end
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.

“‘erb <%= super %> # double-renders <% super %> # doesn’t double-render “‘

‘super` also doesn’t consider the current variant. ‘render_parent` renders the parent template considering the current variant and emits the result without double-rendering.



136
137
138
139
# File 'lib/view_component/base.rb', line 136

def render_parent
  render_parent_to_string
  nil
end

#render_parent_to_stringObject

Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.

“‘ruby def call

"<div>#{render_parent_to_string}</div>"

end “‘

When rendering the parent inside an .erb template, use ‘#render_parent` instead.



151
152
153
154
155
156
157
158
159
160
161
162
# File 'lib/view_component/base.rb', line 151

def render_parent_to_string
  @__vc_parent_render_level ||= 0 # ensure a good starting value

  begin
    target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level]
    @__vc_parent_render_level += 1

    target_render.bind_call(self, @__vc_requested_details)
  ensure
    @__vc_parent_render_level -= 1
  end
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


275
276
277
# File 'lib/view_component/base.rb', line 275

def request
  __vc_request
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



63
64
65
# File 'lib/view_component/base.rb', line 63

def set_original_view_context(view_context)
  self.__vc_original_view_context = view_context
end

#view_cache_dependenciesObject

For caching, such as #cache_if



267
268
269
# File 'lib/view_component/base.rb', line 267

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



261
262
263
# File 'lib/view_component/base.rb', line 261

def virtual_path
  self.class.virtual_path
end