Class: ViewComponent::Base
- Inherits:
-
Object
- Object
- ViewComponent::Base
- Includes:
- ActionView::Helpers, ActiveSupport::CoreExt::ERBUtil, ERB::Escape, InlineTemplate, Slotable, Translatable, WithContentHelper
- Defined in:
- lib/view_component/base.rb
Class Attribute Summary collapse
-
.identifier ⇒ String
The file path of the component Ruby file.
- .virtual_path ⇒ Object
Instance Attribute Summary collapse
-
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
-
#current_template ⇒ Object
readonly
Returns the value of attribute current_template.
Class Method Summary collapse
- .__vc_collection_counter_parameter ⇒ Object
- .__vc_collection_iteration_parameter ⇒ Object
- .__vc_collection_parameter ⇒ Object
- .__vc_compile(raise_errors: false, force: false) ⇒ Object
- .__vc_compiled? ⇒ Boolean
- .__vc_compiler ⇒ Object
- .__vc_counter_argument_present? ⇒ Boolean
- .__vc_ensure_compiled ⇒ Object
- .__vc_iteration_argument_present? ⇒ Boolean
-
.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter.
-
.__vc_validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.
-
.config ⇒ ActiveSupport::OrderedOptions
Returns the current config.
- .inherited(child) ⇒ Object
-
.sidecar_files(extensions) ⇒ Object
Find sidecar files for the given extensions.
-
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
-
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
-
.with_collection(collection, spacer_component: nil, **args) ⇒ Object
Render a component for each element in a collection ([documentation](/guide/collections)):.
-
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):.
Instance Method Summary collapse
- #__vc_request ⇒ Object
-
#around_render ⇒ void
Called around rendering the component.
-
#before_render ⇒ void
Called before rendering the component.
-
#content ⇒ String
The content passed to the component instance as a block.
-
#content? ⇒ Boolean
Whether ‘content` has been passed to the component.
-
#controller ⇒ ActionController::Base
The current controller.
- #format ⇒ Object
-
#helpers ⇒ ActionView::Base
A proxy through which to access helpers.
-
#initialize ⇒ Base
constructor
Including ‘Rails.application.routes.url_helpers` defines an initializer that accepts (…), so we have to define our own empty initializer to overwrite it.
- #method_missing(method_name, *args) ⇒ Object
-
#output_postamble ⇒ String
Optional content to be returned after the rendered template.
-
#output_preamble ⇒ String
Optional content to be returned before the rendered template.
-
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we’re not rendering a component.
-
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
-
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
-
#render_parent ⇒ Object
Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.
-
#render_parent_to_string ⇒ Object
Renders the parent component to a string and returns it.
-
#request ⇒ ActionDispatch::Request
The current request.
-
#set_original_view_context(view_context) ⇒ void
Components render in their own view context.
-
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if.
-
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method.
- #with_original_virtual_path ⇒ Object
Methods included from WithContentHelper
Methods included from Translatable
Methods included from Slotable
#__vc_get_slot, #__vc_set_polymorphic_slot, #__vc_set_slot
Constructor Details
#initialize ⇒ Base
Including ‘Rails.application.routes.url_helpers` defines an initializer that accepts (…), so we have to define our own empty initializer to overwrite it.
96 97 |
# File 'lib/view_component/base.rb', line 96 def initialize end |
Dynamic Method Handling
This class handles dynamic methods through the method_missing method
#method_missing(method_name, *args) ⇒ Object
294 295 296 297 298 299 300 301 302 303 304 305 |
# File 'lib/view_component/base.rb', line 294 def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing super rescue => e # rubocop:disable Style/RescueStandardError e.set_backtrace e.backtrace.tap(&:shift) raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name) #{e.} You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}`? MESSAGE raise end |
Class Attribute Details
.identifier ⇒ String
The file path of the component Ruby file.
481 482 483 |
# File 'lib/view_component/base.rb', line 481 def identifier @identifier end |
.virtual_path ⇒ Object
487 488 489 |
# File 'lib/view_component/base.rb', line 487 def virtual_path @virtual_path end |
Instance Attribute Details
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
76 77 78 |
# File 'lib/view_component/base.rb', line 76 def __vc_original_view_context @__vc_original_view_context end |
#current_template ⇒ Object (readonly)
Returns the value of attribute current_template.
77 78 79 |
# File 'lib/view_component/base.rb', line 77 def current_template @current_template end |
Class Method Details
.__vc_collection_counter_parameter ⇒ Object
670 671 672 |
# File 'lib/view_component/base.rb', line 670 def __vc_collection_counter_parameter @__vc_collection_counter_parameter ||= :"#{__vc_collection_parameter}_counter" end |
.__vc_collection_iteration_parameter ⇒ Object
680 681 682 |
# File 'lib/view_component/base.rb', line 680 def __vc_collection_iteration_parameter @__vc_collection_iteration_parameter ||= :"#{__vc_collection_parameter}_iteration" end |
.__vc_collection_parameter ⇒ Object
665 666 667 |
# File 'lib/view_component/base.rb', line 665 def __vc_collection_parameter @__vc_provided_collection_parameter ||= name && name.demodulize.underscore.chomp("_component").to_sym end |
.__vc_compile(raise_errors: false, force: false) ⇒ Object
545 546 547 |
# File 'lib/view_component/base.rb', line 545 def __vc_compile(raise_errors: false, force: false) __vc_compiler.compile(raise_errors: raise_errors, force: force) end |
.__vc_compiled? ⇒ Boolean
593 594 595 |
# File 'lib/view_component/base.rb', line 593 def __vc_compiled? __vc_compiler.compiled? end |
.__vc_compiler ⇒ Object
603 604 605 |
# File 'lib/view_component/base.rb', line 603 def __vc_compiler @__vc_compiler ||= Compiler.new(self) end |
.__vc_counter_argument_present? ⇒ Boolean
675 676 677 |
# File 'lib/view_component/base.rb', line 675 def __vc_counter_argument_present? __vc_initialize_parameter_names.include?(__vc_collection_counter_parameter) end |
.__vc_ensure_compiled ⇒ Object
598 599 600 |
# File 'lib/view_component/base.rb', line 598 def __vc_ensure_compiled __vc_compile unless __vc_compiled? end |
.__vc_iteration_argument_present? ⇒ Boolean
685 686 687 |
# File 'lib/view_component/base.rb', line 685 def __vc_iteration_argument_present? __vc_initialize_parameter_names.include?(__vc_collection_iteration_parameter) end |
.__vc_validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.
645 646 647 648 649 650 651 652 |
# File 'lib/view_component/base.rb', line 645 def __vc_validate_collection_parameter!(validate_default: false) parameter = validate_default ? __vc_collection_parameter : __vc_provided_collection_parameter return unless parameter return if __vc_initialize_parameter_names.include?(parameter) || __vc_splatted_keyword_argument_present? raise MissingCollectionArgumentError.new(name, parameter) end |
.__vc_validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.
658 659 660 661 662 |
# File 'lib/view_component/base.rb', line 658 def __vc_validate_initialization_parameters! return unless __vc_initialize_parameter_names.include?(:content) raise ReservedParameterError.new(name, :content) end |
.config ⇒ ActiveSupport::OrderedOptions
Returns the current config.
39 40 41 42 43 44 45 46 |
# File 'lib/view_component/base.rb', line 39 def config module_parents.each do |module_parent| next unless module_parent.respond_to?(:config) module_parent_config = module_parent.config.try(:view_component) return module_parent_config if module_parent_config end ViewComponent::Config.current end |
.inherited(child) ⇒ Object
550 551 552 553 554 555 556 557 558 559 560 561 562 563 564 565 566 567 568 569 570 571 572 573 574 575 576 577 578 579 580 581 582 583 584 585 586 587 588 589 590 |
# File 'lib/view_component/base.rb', line 550 def inherited(child) # Compile so child will inherit compiled `call_*` template methods that # `compile` defines __vc_compile # Give the child its own personal #render_template_for to protect against the case when # eager loading is disabled and the parent component is rendered before the child. In # such a scenario, the parent will override ViewComponent::Base#render_template_for, # meaning it will not be called for any children and thus not compile their templates. if !child.instance_methods(false).include?(:render_template_for) && !child.__vc_compiled? child.class_eval <<~RUBY, __FILE__, __LINE__ + 1 def render_template_for(requested_details) # Force compilation here so the compiler always redefines render_template_for. # This is mostly a safeguard to prevent infinite recursion. self.class.__vc_compile(raise_errors: true, force: true) # .__vc_compile replaces this method; call the new one render_template_for(requested_details) end RUBY end # Derive the source location of the component Ruby file from the call stack. # We need to ignore `inherited` frames here as they indicate that `inherited` # has been re-defined by the consuming application, likely in ApplicationComponent. # We use `base_label` method here instead of `label` to avoid cases where the method # owner is included in a prefix like `ApplicationComponent.inherited`. child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path child.virtual_path = child.name&.underscore # Set collection parameter to the extended component child.with_collection_parameter(__vc_provided_collection_parameter) if instance_methods(false).include?(:render_template_for) vc_ancestor_calls = defined?(@__vc_ancestor_calls) ? @__vc_ancestor_calls.dup : [] vc_ancestor_calls.unshift(instance_method(:render_template_for)) child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls) end super end |
.sidecar_files(extensions) ⇒ Object
Find sidecar files for the given extensions.
The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.
For example, one might collect sidecar CSS files that need to be compiled.
496 497 498 499 500 501 502 503 504 505 506 507 508 509 510 511 512 513 514 515 516 517 518 519 520 521 522 523 524 525 526 527 528 529 |
# File 'lib/view_component/base.rb', line 496 def sidecar_files(extensions) return [] unless identifier extensions = extensions.join(",") # view files in a directory named like the component directory = File.dirname(identifier) filename = File.basename(identifier, ".rb") component_name = name.demodulize.underscore # Add support for nested components defined in the same file. # # for example # # class MyComponent < ViewComponent::Base # class MyOtherComponent < ViewComponent::Base # end # end # # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb` nested_component_files = if name.include?("::") && component_name != filename Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"] else [] end # view files in the same directory as the component sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"] sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"] (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq end |
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
“‘ruby class MyComponent < ViewComponent::Base
strip_trailing_whitespace
end “‘
628 629 630 |
# File 'lib/view_component/base.rb', line 628 def strip_trailing_whitespace(value = true) self.__vc_strip_trailing_whitespace = value end |
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
635 636 637 |
# File 'lib/view_component/base.rb', line 635 def strip_trailing_whitespace? __vc_strip_trailing_whitespace end |
.with_collection(collection, spacer_component: nil, **args) ⇒ Object
Render a component for each element in a collection ([documentation](/guide/collections)):
“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`
540 541 542 |
# File 'lib/view_component/base.rb', line 540 def with_collection(collection, spacer_component: nil, **args) Collection.new(self, collection, spacer_component, **args) end |
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):
“‘ruby with_collection_parameter :item “`
614 615 616 617 |
# File 'lib/view_component/base.rb', line 614 def with_collection_parameter(parameter) @__vc_provided_collection_parameter = parameter @__vc_initialize_parameters = nil end |
Instance Method Details
#__vc_request ⇒ Object
330 331 332 333 |
# File 'lib/view_component/base.rb', line 330 def __vc_request # The current request (if present, as mailers/jobs/etc do not have a request) @__vc_request ||= controller.request if controller.respond_to?(:request) end |
#around_render ⇒ void
This method returns an undefined value.
Called around rendering the component. Override to wrap the rendering of a component in custom instrumentation, etc.
238 239 240 |
# File 'lib/view_component/base.rb', line 238 def around_render yield end |
#before_render ⇒ void
This method returns an undefined value.
Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.
230 231 232 |
# File 'lib/view_component/base.rb', line 230 def before_render # noop end |
#content ⇒ String
The content passed to the component instance as a block.
338 339 340 341 342 343 344 345 346 347 348 349 350 |
# File 'lib/view_component/base.rb', line 338 def content @__vc_content_evaluated = true return @__vc_content if defined?(@__vc_content) @__vc_content = if __vc_render_in_block_provided? with_original_virtual_path do view_context.capture(self, &@__vc_render_in_block) end elsif __vc_content_set_by_with_content_defined? @__vc_content_set_by_with_content end end |
#content? ⇒ Boolean
Whether ‘content` has been passed to the component.
355 356 357 |
# File 'lib/view_component/base.rb', line 355 def content? __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined? end |
#controller ⇒ ActionController::Base
The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
269 270 271 272 273 |
# File 'lib/view_component/base.rb', line 269 def controller raise ControllerCalledBeforeRenderError if view_context.nil? @__vc_controller ||= view_context.controller end |
#format ⇒ Object
359 360 361 |
# File 'lib/view_component/base.rb', line 359 def format self.class.__vc_response_format end |
#helpers ⇒ ActionView::Base
A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
279 280 281 282 283 284 285 286 287 288 289 290 |
# File 'lib/view_component/base.rb', line 279 def helpers raise HelpersCalledBeforeRenderError if view_context.nil? # Attempt to re-use the original view_context passed to the first # component rendered in the rendering pipeline. This prevents the # instantiation of a new view_context via `controller.view_context` which # always returns a new instance of the view context class. # # This allows ivars to remain persisted when using the same helper via # `helpers` across multiple components and partials. @__vc_helpers ||= __vc_original_view_context || controller.view_context end |
#output_postamble ⇒ String
Optional content to be returned after the rendered template.
222 223 224 |
# File 'lib/view_component/base.rb', line 222 def output_postamble @@default_output_postamble ||= "".html_safe end |
#output_preamble ⇒ String
Optional content to be returned before the rendered template.
215 216 217 |
# File 'lib/view_component/base.rb', line 215 def output_preamble @@default_output_preamble ||= "".html_safe end |
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we’re not rendering a component.
This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`, and not the component’s ‘view_context`.
256 257 258 259 260 261 262 263 |
# File 'lib/view_component/base.rb', line 256 def render( = {}, args = {}, &block) if .respond_to?(:set_original_view_context) .set_original_view_context(self.__vc_original_view_context) @view_context.render(, args, &block) else __vc_original_view_context.render(, args, &block) end end |
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
245 246 247 |
# File 'lib/view_component/base.rb', line 245 def render? true end |
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
-
‘view_context`: ActionView context from calling view
-
‘block`: optional block to be captured within the view context
Returns HTML that has been escaped by the respective template handler.
107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 |
# File 'lib/view_component/base.rb', line 107 def render_in(view_context, &block) self.class.__vc_compile(raise_errors: true) @view_context = view_context @old_virtual_path = view_context.instance_variable_get(:@virtual_path) self.__vc_original_view_context ||= view_context @output_buffer = view_context.output_buffer @lookup_context ||= view_context.lookup_context # For content_for @view_flow ||= view_context.view_flow # For i18n @virtual_path ||= virtual_path # Describes the inferred request constraints (locales, formats, variants) @__vc_requested_details ||= @lookup_context.vc_requested_details # For caching, such as #cache_if @current_template = nil unless defined?(@current_template) old_current_template = @current_template if block && defined?(@__vc_content_set_by_with_content) raise DuplicateContentError.new(self.class.name) end @__vc_content_evaluated = false @__vc_render_in_block = block before_render if render? value = nil @output_buffer.with_buffer do @view_context.instance_variable_set(:@virtual_path, virtual_path) rendered_template = around_render do render_template_for(@__vc_requested_details).to_s end # Avoid allocating new string when output_preamble and output_postamble are blank value = if output_preamble.blank? && output_postamble.blank? rendered_template else __vc_safe_output_preamble + rendered_template + __vc_safe_output_postamble end end if ActionView::Base.annotate_rendered_view_with_filenames && current_template.inline_call? && request&.format == :html identifier = defined?(Rails.root) ? self.class.identifier.sub("#{Rails.root}/", "") : self.class.identifier value = "<!-- BEGIN #{identifier} -->".html_safe + value + "<!-- END #{identifier} -->".html_safe end value else "" end ensure view_context.instance_variable_set(:@virtual_path, @old_virtual_path) @current_template = old_current_template end |
#render_parent ⇒ Object
Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.
“‘erb <%= super %> # double-renders <% super %> # doesn’t double-render “‘
‘super` also doesn’t consider the current variant. ‘render_parent` renders the parent template considering the current variant and emits the result without double-rendering.
184 185 186 187 |
# File 'lib/view_component/base.rb', line 184 def render_parent render_parent_to_string nil end |
#render_parent_to_string ⇒ Object
Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.
“‘ruby def call
"<div>#{render_parent_to_string}</div>"
end “‘
When rendering the parent inside an .erb template, use ‘#render_parent` instead.
199 200 201 202 203 204 205 206 207 208 209 210 |
# File 'lib/view_component/base.rb', line 199 def render_parent_to_string @__vc_parent_render_level ||= 0 # ensure a good starting value begin target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level] @__vc_parent_render_level += 1 target_render.bind_call(self, @__vc_requested_details) ensure @__vc_parent_render_level -= 1 end end |
#request ⇒ ActionDispatch::Request
The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
325 326 327 |
# File 'lib/view_component/base.rb', line 325 def request __vc_request end |
#set_original_view_context(view_context) ⇒ void
This method returns an undefined value.
Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.
88 89 90 |
# File 'lib/view_component/base.rb', line 88 def set_original_view_context(view_context) # noop end |
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if
317 318 319 |
# File 'lib/view_component/base.rb', line 317 def view_cache_dependencies [] end |
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method
311 312 313 |
# File 'lib/view_component/base.rb', line 311 def virtual_path self.class.virtual_path end |
#with_original_virtual_path ⇒ Object
364 365 366 367 368 369 |
# File 'lib/view_component/base.rb', line 364 def with_original_virtual_path @view_context.instance_variable_set(:@virtual_path, @old_virtual_path) yield ensure @view_context.instance_variable_set(:@virtual_path, virtual_path) end |