Class: ViewComponent::Base

Inherits:
ActionView::Base
  • Object
show all
Includes:
InlineTemplate, Slotable, Translatable, UseHelpers, WithContentHelper
Defined in:
lib/view_component/base.rb

Constant Summary collapse

RESERVED_PARAMETER =
:content
VC_INTERNAL_DEFAULT_FORMAT =
:html

Constants included from Translatable

Translatable::HTML_SAFE_TRANSLATION_KEY, Translatable::TRANSLATION_EXTENSIONS

Constants included from Slotable

Slotable::RESERVED_NAMES

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Constructor Details

#initializeBase

Returns a new instance of Base.



199
200
# File 'lib/view_component/base.rb', line 199

def initialize(*)
end

Dynamic Method Handling

This class handles dynamic methods through the method_missing method

#method_missing(method_name, *args) ⇒ Object



247
248
249
250
251
252
253
254
255
256
257
258
# File 'lib/view_component/base.rb', line 247

def method_missing(method_name, *args) # rubocop:disable Style/MissingRespondToMissing
  super
rescue => e # rubocop:disable Style/RescueStandardError
  e.set_backtrace e.backtrace.tap(&:shift)
  raise e, <<~MESSAGE.chomp if view_context && e.is_a?(NameError) && helpers.respond_to?(method_name)
    #{e.message}

    You may be trying to call a method provided as a view helper. Did you mean `helpers.#{method_name}'?
  MESSAGE

  raise
end

Class Attribute Details

.identifierString

The file path of the component Ruby file.

Returns:

  • (String)


456
457
458
# File 'lib/view_component/base.rb', line 456

def identifier
  @identifier
end

.virtual_pathObject



462
463
464
# File 'lib/view_component/base.rb', line 462

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



55
56
57
# File 'lib/view_component/base.rb', line 55

def __vc_original_view_context
  @__vc_original_view_context
end

Class Method Details

.collection_counter_parameterObject



664
665
666
# File 'lib/view_component/base.rb', line 664

def collection_counter_parameter
  :"#{collection_parameter}_counter"
end

.collection_iteration_parameterObject



674
675
676
# File 'lib/view_component/base.rb', line 674

def collection_iteration_parameter
  :"#{collection_parameter}_iteration"
end

.collection_parameterObject



659
660
661
# File 'lib/view_component/base.rb', line 659

def collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.compile(raise_errors: false, force: false) ⇒ Object



585
586
587
# File 'lib/view_component/base.rb', line 585

def compile(raise_errors: false, force: false)
  compiler.compile(raise_errors: raise_errors, force: force)
end

.compiled?Boolean

Returns:

  • (Boolean)


575
576
577
# File 'lib/view_component/base.rb', line 575

def compiled?
  compiler.compiled?
end

.compilerObject



590
591
592
# File 'lib/view_component/base.rb', line 590

def compiler
  @__vc_compiler ||= Compiler.new(self)
end

.configActiveSupport::OrderedOptions

Returns the current config.

Returns:

  • (ActiveSupport::OrderedOptions)


27
28
29
30
31
32
33
# File 'lib/view_component/base.rb', line 27

def config
  module_parents.each do |m|
    config = m.try(:config).try(:view_component)
    return config if config
  end
  ViewComponent::Config.current
end

.counter_argument_present?Boolean

Returns:

  • (Boolean)


669
670
671
# File 'lib/view_component/base.rb', line 669

def counter_argument_present?
  initialize_parameter_names.include?(collection_counter_parameter)
end

.ensure_compiledObject



580
581
582
# File 'lib/view_component/base.rb', line 580

def ensure_compiled
  compile unless compiled?
end

.inherited(child) ⇒ Object



520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
# File 'lib/view_component/base.rb', line 520

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(variant = nil, format = nil)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.compile(raise_errors: true, force: true)
        # .compile replaces this method; call the new one
        render_template_for(variant, format)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  # We use `base_label` method here instead of `label` to avoid cases where the method
  # owner is included in a prefix like `ApplicationComponent.inherited`.
  child.identifier = caller_locations(1, 10).reject { |l| l.base_label == "inherited" }[0].path

  # If Rails application is loaded, removes the first part of the path and the extension.
  if defined?(Rails) && Rails.application
    child.virtual_path = child.identifier.gsub(
      /(.*#{Regexp.quote(ViewComponent::Base.config.view_component_path)})|(\.rb)/, ""
    )
  end

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  if instance_methods(false).include?(:render_template_for)
    vc_ancestor_calls = defined?(@__vc_ancestor_calls) ? @__vc_ancestor_calls.dup : []

    vc_ancestor_calls.unshift(instance_method(:render_template_for))
    child.instance_variable_set(:@__vc_ancestor_calls, vc_ancestor_calls)
  end

  super
end

.iteration_argument_present?Boolean

Returns:

  • (Boolean)


679
680
681
# File 'lib/view_component/base.rb', line 679

def iteration_argument_present?
  initialize_parameter_names.include?(collection_iteration_parameter)
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
# File 'lib/view_component/base.rb', line 471

def sidecar_files(extensions)
  return [] unless identifier

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(identifier)
  filename = File.basename(identifier, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [identifier] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



615
616
617
# File 'lib/view_component/base.rb', line 615

def strip_trailing_whitespace(value = true)
  self.__vc_strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


622
623
624
# File 'lib/view_component/base.rb', line 622

def strip_trailing_whitespace?
  __vc_strip_trailing_whitespace
end

.validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



632
633
634
635
636
637
638
639
640
641
642
643
644
645
646
# File 'lib/view_component/base.rb', line 632

def validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



652
653
654
655
656
# File 'lib/view_component/base.rb', line 652

def validate_initialization_parameters!
  return unless initialize_parameter_names.include?(RESERVED_PARAMETER)

  raise ReservedParameterError.new(name, RESERVED_PARAMETER)
end

.with_collection(collection, spacer_component: nil, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • spacer_component (ViewComponent::Base) (defaults to: nil)

    Component instance to be rendered between items.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



515
516
517
# File 'lib/view_component/base.rb', line 515

def with_collection(collection, spacer_component: nil, **args)
  Collection.new(self, collection, spacer_component, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



601
602
603
604
# File 'lib/view_component/base.rb', line 601

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
  @initialize_parameters = nil
end

Instance Method Details

#__vc_requestObject

Enables consumers to override request/@request



292
293
294
# File 'lib/view_component/base.rb', line 292

def __vc_request
  @__vc_request ||= controller.request if controller.respond_to?(:request)
end

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



187
188
189
# File 'lib/view_component/base.rb', line 187

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


299
300
301
302
303
304
305
306
307
308
309
# File 'lib/view_component/base.rb', line 299

def content
  @__vc_content_evaluated = true
  return @__vc_content if defined?(@__vc_content)

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content_defined?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


314
315
316
# File 'lib/view_component/base.rb', line 314

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



222
223
224
225
226
# File 'lib/view_component/base.rb', line 222

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#formatObject

For caching, such as #cache_if



277
278
279
# File 'lib/view_component/base.rb', line 277

def format
  @__vc_variant if defined?(@__vc_variant)
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



232
233
234
235
236
237
238
239
240
241
242
243
# File 'lib/view_component/base.rb', line 232

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


179
180
181
# File 'lib/view_component/base.rb', line 179

def output_postamble
  @@default_output_postamble ||= "".html_safe
end

#output_preambleString

Optional content to be returned before the rendered template.

Returns:

  • (String)


172
173
174
# File 'lib/view_component/base.rb', line 172

def output_preamble
  @@default_output_preamble ||= "".html_safe
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`,

and not the component's `view_context`.


209
210
211
212
213
214
215
216
# File 'lib/view_component/base.rb', line 209

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    super
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


194
195
196
# File 'lib/view_component/base.rb', line 194

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
# File 'lib/view_component/base.rb', line 78

def render_in(view_context, &block)
  self.class.compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = ActionView::OutputBuffer.new

  @lookup_context ||= view_context.lookup_context

  # required for path helpers in older Rails versions
  @view_renderer ||= view_context.view_renderer

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # For template variants (+phone, +desktop, etc.)
  @__vc_variant ||= @lookup_context.variants.first

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template
  @current_template = self

  if block && defined?(@__vc_content_set_by_with_content)
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    rendered_template = render_template_for(@__vc_variant, __vc_request&.format&.to_sym).to_s

    # Avoid allocating new string when output_preamble and output_postamble are blank
    if output_preamble.blank? && output_postamble.blank?
      rendered_template
    else
      safe_output_preamble + rendered_template + safe_output_postamble
    end
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result.

“‘erb <%= super %> # double-renders <% super %> # doesn’t double-render “‘

‘super` also doesn’t consider the current variant. ‘render_parent` renders the parent template considering the current variant and emits the result without double-rendering.



141
142
143
144
# File 'lib/view_component/base.rb', line 141

def render_parent
  render_parent_to_string
  nil
end

#render_parent_to_stringObject

Renders the parent component to a string and returns it. This method is meant to be used inside custom #call methods when a string result is desired, eg.

“‘ruby def call

"<div>#{render_parent_to_string}</div>"

end “‘

When rendering the parent inside an .erb template, use ‘#render_parent` instead.



156
157
158
159
160
161
162
163
164
165
166
167
# File 'lib/view_component/base.rb', line 156

def render_parent_to_string
  @__vc_parent_render_level ||= 0 # ensure a good starting value

  begin
    target_render = self.class.instance_variable_get(:@__vc_ancestor_calls)[@__vc_parent_render_level]
    @__vc_parent_render_level += 1

    target_render.bind_call(self, @__vc_variant)
  ensure
    @__vc_parent_render_level -= 1
  end
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


285
286
287
# File 'lib/view_component/base.rb', line 285

def request
  __vc_request
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



66
67
68
# File 'lib/view_component/base.rb', line 66

def set_original_view_context(view_context)
  self.__vc_original_view_context = view_context
end

#view_cache_dependenciesObject

For caching, such as #cache_if



270
271
272
# File 'lib/view_component/base.rb', line 270

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



264
265
266
# File 'lib/view_component/base.rb', line 264

def virtual_path
  self.class.virtual_path
end