Class: ViewComponent::Base

Inherits:
ActionView::Base
  • Object
show all
Includes:
InlineTemplate, Slotable, Translatable, WithContentHelper
Defined in:
lib/view_component/base.rb

Constant Summary collapse

RESERVED_PARAMETER =
:content

Constants included from Translatable

Translatable::HTML_SAFE_TRANSLATION_KEY

Constants included from Slotable

Slotable::RESERVED_NAMES

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Constructor Details

#initializeBase

Returns a new instance of Base.



156
157
# File 'lib/view_component/base.rb', line 156

def initialize(*)
end

Class Attribute Details

.configActiveSupport::OrderedOptions

Returns the current config.

Returns:

  • (ActiveSupport::OrderedOptions)


24
25
26
# File 'lib/view_component/base.rb', line 24

def config
  @config ||= ActiveSupport::OrderedOptions.new
end

.source_locationObject



356
357
358
# File 'lib/view_component/base.rb', line 356

def source_location
  @source_location
end

.virtual_pathObject



356
357
358
# File 'lib/view_component/base.rb', line 356

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



50
51
52
# File 'lib/view_component/base.rb', line 50

def __vc_original_view_context
  @__vc_original_view_context
end

Class Method Details

.collection_counter_parameterObject



574
575
576
# File 'lib/view_component/base.rb', line 574

def collection_counter_parameter
  "#{collection_parameter}_counter".to_sym
end

.collection_iteration_parameterObject



584
585
586
# File 'lib/view_component/base.rb', line 584

def collection_iteration_parameter
  "#{collection_parameter}_iteration".to_sym
end

.collection_parameterObject



569
570
571
# File 'lib/view_component/base.rb', line 569

def collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.compile(raise_errors: false, force: false) ⇒ Object

Compile templates to instance methods, assuming they haven’t been compiled already.

Do as much work as possible in this step, as doing so reduces the amount of work done each time a component is rendered.



480
481
482
# File 'lib/view_component/base.rb', line 480

def compile(raise_errors: false, force: false)
  compiler.compile(raise_errors: raise_errors, force: force)
end

.compiled?Boolean

Returns:

  • (Boolean)


466
467
468
# File 'lib/view_component/base.rb', line 466

def compiled?
  compiler.compiled?
end

.compilerObject



485
486
487
# File 'lib/view_component/base.rb', line 485

def compiler
  @__vc_compiler ||= Compiler.new(self)
end

.counter_argument_present?Boolean

Returns:

  • (Boolean)


579
580
581
# File 'lib/view_component/base.rb', line 579

def counter_argument_present?
  initialize_parameter_names.include?(collection_counter_parameter)
end

.ensure_compiledObject



471
472
473
# File 'lib/view_component/base.rb', line 471

def ensure_compiled
  compile unless compiled?
end

.formatObject



496
497
498
# File 'lib/view_component/base.rb', line 496

def format
  :html
end

.identifierObject



501
502
503
# File 'lib/view_component/base.rb', line 501

def identifier
  source_location
end

.inherited(child) ⇒ Object



420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
# File 'lib/view_component/base.rb', line 420

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(variant = nil)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.compile(raise_errors: true, force: true)
        # .compile replaces this method; call the new one
        render_template_for(variant)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  child.source_location = caller_locations(1, 10).reject { |l| l.label == "inherited" }[0].path

  # If Rails application is loaded, removes the first part of the path and the extension.
  if defined?(Rails) && Rails.application
    child.virtual_path = child.source_location.gsub(
      /(.*#{Regexp.quote(ViewComponent::Base.config.view_component_path)})|(\.rb)/, ""
    )
  end

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  super
end

.iteration_argument_present?Boolean

Returns:

  • (Boolean)


589
590
591
# File 'lib/view_component/base.rb', line 589

def iteration_argument_present?
  initialize_parameter_names.include?(collection_iteration_parameter)
end

.short_identifierObject

Provide identifier for ActionView template annotations



415
416
417
# File 'lib/view_component/base.rb', line 415

def short_identifier
  @short_identifier ||= defined?(Rails.root) ? source_location.sub("#{Rails.root}/", "") : source_location
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
# File 'lib/view_component/base.rb', line 365

def sidecar_files(extensions)
  return [] unless source_location

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(source_location)
  filename = File.basename(source_location, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [source_location] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



525
526
527
# File 'lib/view_component/base.rb', line 525

def strip_trailing_whitespace(value = true)
  self.__vc_strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


532
533
534
# File 'lib/view_component/base.rb', line 532

def strip_trailing_whitespace?
  __vc_strip_trailing_whitespace
end

.typeObject

we’ll eventually want to update this to support other types



491
492
493
# File 'lib/view_component/base.rb', line 491

def type
  "text/html"
end

.validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
# File 'lib/view_component/base.rb', line 542

def validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



562
563
564
565
566
# File 'lib/view_component/base.rb', line 562

def validate_initialization_parameters!
  return unless initialize_parameter_names.include?(RESERVED_PARAMETER)

  raise ReservedParameterError.new(name, RESERVED_PARAMETER)
end

.with_collection(collection, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



408
409
410
# File 'lib/view_component/base.rb', line 408

def with_collection(collection, **args)
  Collection.new(self, collection, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



512
513
514
# File 'lib/view_component/base.rb', line 512

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
end

Instance Method Details

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



144
145
146
# File 'lib/view_component/base.rb', line 144

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


233
234
235
236
237
238
239
240
241
242
243
# File 'lib/view_component/base.rb', line 233

def content
  @__vc_content_evaluated = true
  return @__vc_content if defined?(@__vc_content)

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content_defined?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


248
249
250
# File 'lib/view_component/base.rb', line 248

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



179
180
181
182
183
# File 'lib/view_component/base.rb', line 179

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#formatObject

For caching, such as #cache_if



218
219
220
# File 'lib/view_component/base.rb', line 218

def format
  @__vc_variant if defined?(@__vc_variant)
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



189
190
191
192
193
194
195
196
197
198
199
200
# File 'lib/view_component/base.rb', line 189

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


136
137
138
# File 'lib/view_component/base.rb', line 136

def output_postamble
  ""
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`,

and not the component's `view_context`.


166
167
168
169
170
171
172
173
# File 'lib/view_component/base.rb', line 166

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    super
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


151
152
153
# File 'lib/view_component/base.rb', line 151

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
# File 'lib/view_component/base.rb', line 73

def render_in(view_context, &block)
  self.class.compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = ActionView::OutputBuffer.new

  @lookup_context ||= view_context.lookup_context

  # required for path helpers in older Rails versions
  @view_renderer ||= view_context.view_renderer

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # For template variants (+phone, +desktop, etc.)
  @__vc_variant ||= @lookup_context.variants.first

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template
  @current_template = self

  if block && defined?(@__vc_content_set_by_with_content)
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    render_template_for(@__vc_variant).to_s + output_postamble
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result:

“‘erb <%= super %> # double-renders <% super %> # does not double-render “`

Calls ‘super`, returning `nil` to avoid rendering the result twice.



127
128
129
130
131
# File 'lib/view_component/base.rb', line 127

def render_parent
  mtd = @__vc_variant ? "call_#{@__vc_variant}" : "call"
  method(mtd).super_method.call
  nil
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


226
227
228
# File 'lib/view_component/base.rb', line 226

def request
  @request ||= controller.request if controller.respond_to?(:request)
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



61
62
63
# File 'lib/view_component/base.rb', line 61

def set_original_view_context(view_context)
  self.__vc_original_view_context = view_context
end

#view_cache_dependenciesObject

For caching, such as #cache_if



211
212
213
# File 'lib/view_component/base.rb', line 211

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



205
206
207
# File 'lib/view_component/base.rb', line 205

def virtual_path
  self.class.virtual_path
end