Class: ViewComponent::Base

Inherits:
ActionView::Base
  • Object
show all
Includes:
Slotable, Translatable, WithContentHelper
Defined in:
lib/view_component/base.rb

Constant Summary collapse

RESERVED_PARAMETER =
:content

Constants included from Translatable

Translatable::HTML_SAFE_TRANSLATION_KEY

Constants included from Slotable

Slotable::RESERVED_NAMES

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#i18n_scope, #translate

Methods included from Slotable

#get_slot, #set_polymorphic_slot, #set_slot

Constructor Details

#initializeBase

Returns a new instance of Base.



151
152
# File 'lib/view_component/base.rb', line 151

def initialize(*)
end

Class Attribute Details

.configViewComponent::Config

Returns the current config.



23
24
25
# File 'lib/view_component/base.rb', line 23

def config
  @config ||= ActiveSupport::OrderedOptions.new
end

.source_locationObject



351
352
353
# File 'lib/view_component/base.rb', line 351

def source_location
  @source_location
end

.virtual_pathObject



351
352
353
# File 'lib/view_component/base.rb', line 351

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



45
46
47
# File 'lib/view_component/base.rb', line 45

def __vc_original_view_context
  @__vc_original_view_context
end

Class Method Details

.collection_counter_parameterObject



567
568
569
# File 'lib/view_component/base.rb', line 567

def collection_counter_parameter
  "#{collection_parameter}_counter".to_sym
end

.collection_iteration_parameterObject



577
578
579
# File 'lib/view_component/base.rb', line 577

def collection_iteration_parameter
  "#{collection_parameter}_iteration".to_sym
end

.collection_parameterObject



562
563
564
# File 'lib/view_component/base.rb', line 562

def collection_parameter
  provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym
end

.compile(raise_errors: false, force: false) ⇒ Object

Compile templates to instance methods, assuming they haven’t been compiled already.

Do as much work as possible in this step, as doing so reduces the amount of work done each time a component is rendered.



473
474
475
# File 'lib/view_component/base.rb', line 473

def compile(raise_errors: false, force: false)
  compiler.compile(raise_errors: raise_errors, force: force)
end

.compiled?Boolean

Returns:

  • (Boolean)


459
460
461
# File 'lib/view_component/base.rb', line 459

def compiled?
  compiler.compiled?
end

.compilerObject



478
479
480
# File 'lib/view_component/base.rb', line 478

def compiler
  @__vc_compiler ||= Compiler.new(self)
end

.counter_argument_present?Boolean

Returns:

  • (Boolean)


572
573
574
# File 'lib/view_component/base.rb', line 572

def counter_argument_present?
  initialize_parameter_names.include?(collection_counter_parameter)
end

.ensure_compiledObject



464
465
466
# File 'lib/view_component/base.rb', line 464

def ensure_compiled
  compile unless compiled?
end

.formatObject



489
490
491
# File 'lib/view_component/base.rb', line 489

def format
  :html
end

.identifierObject



494
495
496
# File 'lib/view_component/base.rb', line 494

def identifier
  source_location
end

.inherited(child) ⇒ Object



415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
# File 'lib/view_component/base.rb', line 415

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(variant = nil)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.compile(raise_errors: true, force: true)
        # .compile replaces this method; call the new one
        render_template_for(variant)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers)
    child.include Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  child.source_location = caller_locations(1, 10).reject { |l| l.label == "inherited" }[0].path

  # Removes the first part of the path and the extension.
  child.virtual_path = child.source_location.gsub(
    /(.*#{Regexp.quote(ViewComponent::Base.config.view_component_path)})|(\.rb)/, ""
  )

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  super
end

.iteration_argument_present?Boolean

Returns:

  • (Boolean)


582
583
584
# File 'lib/view_component/base.rb', line 582

def iteration_argument_present?
  initialize_parameter_names.include?(collection_iteration_parameter)
end

.short_identifierObject

Provide identifier for ActionView template annotations



410
411
412
# File 'lib/view_component/base.rb', line 410

def short_identifier
  @short_identifier ||= defined?(Rails.root) ? source_location.sub("#{Rails.root}/", "") : source_location
end

.sidecar_files(extensions) ⇒ Object

Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the dot, example: ‘[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.

Parameters:

  • extensions (Array<String>)

    Extensions of which to return matching sidecar files.



360
361
362
363
364
365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
# File 'lib/view_component/base.rb', line 360

def sidecar_files(extensions)
  return [] unless source_location

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(source_location)
  filename = File.basename(source_location, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [source_location] + sidecar_directory_files + nested_component_files).uniq
end

.strip_trailing_whitespace(value = true) ⇒ Object

Strips trailing whitespace from templates before compiling them.

“‘ruby class MyComponent < ViewComponent::Base

strip_trailing_whitespace

end “‘

Parameters:

  • value (Boolean) (defaults to: true)

    Whether to strip newlines.



518
519
520
# File 'lib/view_component/base.rb', line 518

def strip_trailing_whitespace(value = true)
  self.__vc_strip_trailing_whitespace = value
end

.strip_trailing_whitespace?Boolean

Whether trailing whitespace will be stripped before compilation.

Returns:

  • (Boolean)


525
526
527
# File 'lib/view_component/base.rb', line 525

def strip_trailing_whitespace?
  __vc_strip_trailing_whitespace
end

.typeObject

we’ll eventually want to update this to support other types



484
485
486
# File 'lib/view_component/base.rb', line 484

def type
  "text/html"
end

.validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don’t validate that the default parameter name is accepted, as support for collection rendering is optional.



535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
# File 'lib/view_component/base.rb', line 535

def validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter) || splatted_keyword_argument_present?

  # If Ruby can't parse the component class, then the initialize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise EmptyOrInvalidInitializerError.new(name, parameter)
  end

  raise MissingCollectionArgumentError.new(name, parameter)
end

.validate_initialization_parameters!Object

Ensure the component initializer doesn’t define invalid parameters that could override the framework’s methods.



555
556
557
558
559
# File 'lib/view_component/base.rb', line 555

def validate_initialization_parameters!
  return unless initialize_parameter_names.include?(RESERVED_PARAMETER)

  raise ReservedParameterError.new(name, RESERVED_PARAMETER)
end

.with_collection(collection, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

“‘ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



403
404
405
# File 'lib/view_component/base.rb', line 403

def with_collection(collection, **args)
  Collection.new(self, collection, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

“‘ruby with_collection_parameter :item “`

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



505
506
507
# File 'lib/view_component/base.rb', line 505

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
end

Instance Method Details

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



139
140
141
# File 'lib/view_component/base.rb', line 139

def before_render
  # noop
end

#contentString

The content passed to the component instance as a block.

Returns:

  • (String)


228
229
230
231
232
233
234
235
236
237
238
# File 'lib/view_component/base.rb', line 228

def content
  @__vc_content_evaluated = true
  return @__vc_content if defined?(@__vc_content)

  @__vc_content =
    if __vc_render_in_block_provided?
      view_context.capture(self, &@__vc_render_in_block)
    elsif __vc_content_set_by_with_content_defined?
      @__vc_content_set_by_with_content
    end
end

#content?Boolean

Whether ‘content` has been passed to the component.

Returns:

  • (Boolean)


243
244
245
# File 'lib/view_component/base.rb', line 243

def content?
  __vc_render_in_block_provided? || __vc_content_set_by_with_content_defined?
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)

Raises:



174
175
176
177
178
# File 'lib/view_component/base.rb', line 174

def controller
  raise ControllerCalledBeforeRenderError if view_context.nil?

  @__vc_controller ||= view_context.controller
end

#formatObject

For caching, such as #cache_if



213
214
215
# File 'lib/view_component/base.rb', line 213

def format
  @__vc_variant if defined?(@__vc_variant)
end

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)

Raises:



184
185
186
187
188
189
190
191
192
193
194
195
# File 'lib/view_component/base.rb', line 184

def helpers
  raise HelpersCalledBeforeRenderError if view_context.nil?

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#output_postambleString

Optional content to be returned after the rendered template.

Returns:

  • (String)


131
132
133
# File 'lib/view_component/base.rb', line 131

def output_postamble
  ""
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we’re not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent ‘view_context`,

and not the component's `view_context`.


161
162
163
164
165
166
167
168
# File 'lib/view_component/base.rb', line 161

def render(options = {}, args = {}, &block)
  if options.respond_to?(:set_original_view_context)
    options.set_original_view_context(self.__vc_original_view_context)
    super
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


146
147
148
# File 'lib/view_component/base.rb', line 146

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • ‘view_context`: ActionView context from calling view

  • ‘block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
# File 'lib/view_component/base.rb', line 68

def render_in(view_context, &block)
  self.class.compile(raise_errors: true)

  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = ActionView::OutputBuffer.new

  @lookup_context ||= view_context.lookup_context

  # required for path helpers in older Rails versions
  @view_renderer ||= view_context.view_renderer

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # For template variants (+phone, +desktop, etc.)
  @__vc_variant ||= @lookup_context.variants.first

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template
  @current_template = self

  if block && defined?(@__vc_content_set_by_with_content)
    raise DuplicateContentError.new(self.class.name)
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    render_template_for(@__vc_variant).to_s + output_postamble
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call ‘super` inside their template code will cause a double render if they emit the result:

“‘erb <%= super %> # double-renders <% super %> # does not double-render “`

Calls ‘super`, returning `nil` to avoid rendering the result twice.



122
123
124
125
126
# File 'lib/view_component/base.rb', line 122

def render_parent
  mtd = @__vc_variant ? "call_#{@__vc_variant}" : "call"
  method(mtd).super_method.call
  nil
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


221
222
223
# File 'lib/view_component/base.rb', line 221

def request
  @request ||= controller.request if controller.respond_to?(:request)
end

#set_original_view_context(view_context) ⇒ void

This method returns an undefined value.

Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of ‘ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component’s view context, while objects that don’t will render in the original view context so helpers, etc work as expected.

Parameters:

  • view_context (ActionView::Base)

    The original view context.



56
57
58
# File 'lib/view_component/base.rb', line 56

def set_original_view_context(view_context)
  self.__vc_original_view_context = view_context
end

#view_cache_dependenciesObject

For caching, such as #cache_if



206
207
208
# File 'lib/view_component/base.rb', line 206

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



200
201
202
# File 'lib/view_component/base.rb', line 200

def virtual_path
  self.class.virtual_path
end