Class: ViewComponent::Base
- Inherits:
-
ActionView::Base
- Object
- ActionView::Base
- ViewComponent::Base
- Includes:
- ContentAreas, PolymorphicSlots, SlotableV2, Translatable, WithContentHelper
- Defined in:
- lib/view_component/base.rb
Direct Known Subclasses
Constant Summary collapse
- ViewContextCalledBeforeRenderError =
Class.new(StandardError)
- RESERVED_PARAMETER =
:content
Constants included from Translatable
Translatable::HTML_SAFE_TRANSLATION_KEY
Constants included from SlotableV2
Class Attribute Summary collapse
Instance Attribute Summary collapse
-
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
Class Method Summary collapse
- ._deprecated_generate_mattr_accessor(name) ⇒ Object
-
._sidecar_files(extensions) ⇒ Object
EXPERIMENTAL: This API is experimental and may be removed at any time.
- .collection_counter_parameter ⇒ Object
- .collection_iteration_parameter ⇒ Object
- .collection_parameter ⇒ Object
-
.compile(raise_errors: false, force: false) ⇒ Object
Compile templates to instance methods, assuming they haven't been compiled already.
- .compiled? ⇒ Boolean
- .compiler ⇒ Object
- .config ⇒ Object
- .counter_argument_present? ⇒ Boolean
- .format ⇒ Object
- .identifier ⇒ Object
- .inherited(child) ⇒ Object
- .iteration_argument_present? ⇒ Boolean
-
.short_identifier ⇒ Object
Provide identifier for ActionView template annotations.
-
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
-
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
-
.type ⇒ Object
we'll eventually want to update this to support other types.
-
.validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter.
-
.validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn't define invalid parameters that could override the framework's methods.
-
.with_collection(collection, **args) ⇒ Object
Render a component for each element in a collection ([documentation](/guide/collections)):.
-
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):.
Instance Method Summary collapse
-
#_output_postamble ⇒ String
EXPERIMENTAL: Optional content to be returned after the rendered template.
-
#before_render ⇒ void
Called before rendering the component.
-
#before_render_check ⇒ void
deprecated
Deprecated.
Use `#before_render` instead. Will be removed in v3.0.0.
-
#controller ⇒ ActionController::Base
The current controller.
-
#format ⇒ Object
For caching, such as #cache_if.
-
#generate_distinct_locale_files ⇒ Object
deprecated
Deprecated.
Use `#generate.distinct_locale_files` instead. Will be removed in v3.0.0.
-
#generate_locale ⇒ Object
deprecated
Deprecated.
Use `#generate.locale` instead. Will be removed in v3.0.0.
-
#generate_sidecar ⇒ Object
deprecated
Deprecated.
Use `#generate.sidecar` instead. Will be removed in v3.0.0.
-
#generate_stimulus_controller ⇒ Object
deprecated
Deprecated.
Use `#generate.stimulus_controller` instead. Will be removed in v3.0.0.
-
#helpers ⇒ ActionView::Base
A proxy through which to access helpers.
-
#initialize ⇒ Base
constructor
A new instance of Base.
-
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we're not rendering a component.
-
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
-
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
-
#render_parent ⇒ Object
Subclass components that call `super` inside their template code will cause a double render if they emit the result:.
-
#request ⇒ ActionDispatch::Request
The current request.
-
#set_original_view_context(view_context) ⇒ void
Components render in their own view context.
-
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if.
-
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method.
-
#with_variant(variant) ⇒ self
deprecated
Deprecated.
Will be removed in v3.0.0.
Methods included from WithContentHelper
Methods included from Translatable
#html_safe_translation, #i18n_scope, #translate
Methods included from SlotableV2
Methods included from PolymorphicSlots
Methods included from ContentAreas
Constructor Details
#initialize ⇒ Base
Returns a new instance of Base.
185 186 |
# File 'lib/view_component/base.rb', line 185 def initialize(*) end |
Class Attribute Details
.source_location ⇒ Object
410 411 412 |
# File 'lib/view_component/base.rb', line 410 def source_location @source_location end |
.virtual_path ⇒ Object
410 411 412 |
# File 'lib/view_component/base.rb', line 410 def virtual_path @virtual_path end |
Instance Attribute Details
#__vc_original_view_context ⇒ Object
Returns the value of attribute __vc_original_view_context.
47 48 49 |
# File 'lib/view_component/base.rb', line 47 def __vc_original_view_context @__vc_original_view_context end |
Class Method Details
._deprecated_generate_mattr_accessor(name) ⇒ Object
65 66 67 68 69 70 71 72 |
# File 'lib/view_component/base.rb', line 65 def self._deprecated_generate_mattr_accessor(name) define_singleton_method("generate_#{name}".to_sym) do generate.public_send(name) end define_singleton_method("generate_#{name}=".to_sym) do |value| generate.public_send("#{name}=".to_sym, value) end end |
._sidecar_files(extensions) ⇒ Object
EXPERIMENTAL: This API is experimental and may be removed at any time. Find sidecar files for the given extensions.
The provided array of extensions is expected to contain strings starting without the “dot”, example: `[“erb”, “haml”]`.
For example, one might collect sidecar CSS files that need to be compiled.
420 421 422 423 424 425 426 427 428 429 430 431 432 433 434 435 436 437 438 439 440 441 442 443 444 445 446 447 448 449 450 451 452 453 |
# File 'lib/view_component/base.rb', line 420 def _sidecar_files(extensions) return [] unless source_location extensions = extensions.join(",") # view files in a directory named like the component directory = File.dirname(source_location) filename = File.basename(source_location, ".rb") component_name = name.demodulize.underscore # Add support for nested components defined in the same file. # # for example # # class MyComponent < ViewComponent::Base # class MyOtherComponent < ViewComponent::Base # end # end # # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb` nested_component_files = if name.include?("::") && component_name != filename Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"] else [] end # view files in the same directory as the component sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"] sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"] (sidecar_files - [source_location] + sidecar_directory_files + nested_component_files).uniq end |
.collection_counter_parameter ⇒ Object
629 630 631 |
# File 'lib/view_component/base.rb', line 629 def collection_counter_parameter "#{collection_parameter}_counter".to_sym end |
.collection_iteration_parameter ⇒ Object
639 640 641 |
# File 'lib/view_component/base.rb', line 639 def collection_iteration_parameter "#{collection_parameter}_iteration".to_sym end |
.collection_parameter ⇒ Object
624 625 626 |
# File 'lib/view_component/base.rb', line 624 def collection_parameter provided_collection_parameter || name && name.demodulize.underscore.chomp("_component").to_sym end |
.compile(raise_errors: false, force: false) ⇒ Object
Compile templates to instance methods, assuming they haven't been compiled already.
Do as much work as possible in this step, as doing so reduces the amount of work done each time a component is rendered.
528 529 530 |
# File 'lib/view_component/base.rb', line 528 def compile(raise_errors: false, force: false) compiler.compile(raise_errors: raise_errors, force: force) end |
.compiled? ⇒ Boolean
519 520 521 |
# File 'lib/view_component/base.rb', line 519 def compiled? compiler.compiled? end |
.compiler ⇒ Object
533 534 535 |
# File 'lib/view_component/base.rb', line 533 def compiler @__vc_compiler ||= Compiler.new(self) end |
.config ⇒ Object
22 23 24 |
# File 'lib/view_component/base.rb', line 22 def config @config ||= ViewComponent::Config.defaults end |
.counter_argument_present? ⇒ Boolean
634 635 636 |
# File 'lib/view_component/base.rb', line 634 def counter_argument_present? initialize_parameter_names.include?(collection_counter_parameter) end |
.format ⇒ Object
544 545 546 |
# File 'lib/view_component/base.rb', line 544 def format :html end |
.identifier ⇒ Object
549 550 551 |
# File 'lib/view_component/base.rb', line 549 def identifier source_location end |
.inherited(child) ⇒ Object
475 476 477 478 479 480 481 482 483 484 485 486 487 488 489 490 491 492 493 494 495 496 497 498 499 500 501 502 503 504 505 506 507 508 509 510 511 512 513 514 515 516 |
# File 'lib/view_component/base.rb', line 475 def inherited(child) # Compile so child will inherit compiled `call_*` template methods that # `compile` defines compile # Give the child its own personal #render_template_for to protect against the case when # eager loading is disabled and the parent component is rendered before the child. In # such a scenario, the parent will override ViewComponent::Base#render_template_for, # meaning it will not be called for any children and thus not compile their templates. if !child.instance_methods(false).include?(:render_template_for) && !child.compiled? child.class_eval <<~RUBY, __FILE__, __LINE__ + 1 def render_template_for(variant = nil) # Force compilation here so the compiler always redefines render_template_for. # This is mostly a safeguard to prevent infinite recursion. self.class.compile(raise_errors: true, force: true) # .compile replaces this method; call the new one render_template_for(variant) end RUBY end # If Rails application is loaded, add application url_helpers to the component context # we need to check this to use this gem as a dependency if defined?(Rails) && Rails.application && !(child < Rails.application.routes.url_helpers) child.include Rails.application.routes.url_helpers end # Derive the source location of the component Ruby file from the call stack. # We need to ignore `inherited` frames here as they indicate that `inherited` # has been re-defined by the consuming application, likely in ApplicationComponent. child.source_location = caller_locations(1, 10).reject { |l| l.label == "inherited" }[0].path # Removes the first part of the path and the extension. child.virtual_path = child.source_location.gsub( /(.*#{Regexp.quote(ViewComponent::Base.config.view_component_path)})|(\.rb)/, "" ) # Set collection parameter to the extended component child.with_collection_parameter provided_collection_parameter super end |
.iteration_argument_present? ⇒ Boolean
644 645 646 |
# File 'lib/view_component/base.rb', line 644 def iteration_argument_present? initialize_parameter_names.include?(collection_iteration_parameter) end |
.short_identifier ⇒ Object
Provide identifier for ActionView template annotations
470 471 472 |
# File 'lib/view_component/base.rb', line 470 def short_identifier @short_identifier ||= defined?(Rails.root) ? source_location.sub("#{Rails.root}/", "") : source_location end |
.strip_trailing_whitespace(value = true) ⇒ Object
Strips trailing whitespace from templates before compiling them.
“`ruby class MyComponent < ViewComponent::Base
strip_trailing_whitespace
end “`
573 574 575 |
# File 'lib/view_component/base.rb', line 573 def strip_trailing_whitespace(value = true) self.__vc_strip_trailing_whitespace = value end |
.strip_trailing_whitespace? ⇒ Boolean
Whether trailing whitespace will be stripped before compilation.
580 581 582 |
# File 'lib/view_component/base.rb', line 580 def strip_trailing_whitespace? __vc_strip_trailing_whitespace end |
.type ⇒ Object
we'll eventually want to update this to support other types
539 540 541 |
# File 'lib/view_component/base.rb', line 539 def type "text/html" end |
.validate_collection_parameter!(validate_default: false) ⇒ Object
Ensure the component initializer accepts the collection parameter. By default, we don't validate that the default parameter name is accepted, as support for collection rendering is optional.
590 591 592 593 594 595 596 597 598 599 600 601 602 603 604 605 606 607 608 609 610 |
# File 'lib/view_component/base.rb', line 590 def validate_collection_parameter!(validate_default: false) parameter = validate_default ? collection_parameter : provided_collection_parameter return unless parameter return if initialize_parameter_names.include?(parameter) # If Ruby can't parse the component class, then the initalize # parameters will be empty and ViewComponent will not be able to render # the component. if initialize_parameters.empty? raise ArgumentError, "The #{self} initializer is empty or invalid." \ "It must accept the parameter `#{parameter}` to render it as a collection.\n\n" \ "To fix this issue, update the initializer to accept `#{parameter}`.\n\n" \ "See https://viewcomponent.org/guide/collections.html for more information on rendering collections." end raise ArgumentError, "The initializer for #{self} doesn't accept the parameter `#{parameter}`, " \ "which is required in order to render it as a collection.\n\n" \ "To fix this issue, update the initializer to accept `#{parameter}`.\n\n" \ "See https://viewcomponent.org/guide/collections.html for more information on rendering collections." end |
.validate_initialization_parameters! ⇒ Object
Ensure the component initializer doesn't define invalid parameters that could override the framework's methods.
616 617 618 619 620 621 |
# File 'lib/view_component/base.rb', line 616 def validate_initialization_parameters! return unless initialize_parameter_names.include?(RESERVED_PARAMETER) raise ViewComponent::ComponentError, "#{self} initializer can't accept the parameter `#{RESERVED_PARAMETER}`, as it will override a " \ "public ViewComponent method. To fix this issue, rename the parameter." end |
.with_collection(collection, **args) ⇒ Object
Render a component for each element in a collection ([documentation](/guide/collections)):
“`ruby render(ProductsComponent.with_collection(@products, foo: :bar)) “`
463 464 465 |
# File 'lib/view_component/base.rb', line 463 def with_collection(collection, **args) Collection.new(self, collection, **args) end |
.with_collection_parameter(parameter) ⇒ Object
Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):
“`ruby with_collection_parameter :item “`
560 561 562 |
# File 'lib/view_component/base.rb', line 560 def with_collection_parameter(parameter) @provided_collection_parameter = parameter end |
Instance Method Details
#_output_postamble ⇒ String
EXPERIMENTAL: Optional content to be returned after the rendered template.
157 158 159 |
# File 'lib/view_component/base.rb', line 157 def _output_postamble "" end |
#before_render ⇒ void
This method returns an undefined value.
Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.
165 166 167 |
# File 'lib/view_component/base.rb', line 165 def before_render before_render_check end |
#before_render_check ⇒ void
Use `#before_render` instead. Will be removed in v3.0.0.
This method returns an undefined value.
Called after rendering the component.
173 174 175 |
# File 'lib/view_component/base.rb', line 173 def before_render_check # noop end |
#controller ⇒ ActionController::Base
The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
208 209 210 211 212 213 214 215 216 217 218 219 220 221 |
# File 'lib/view_component/base.rb', line 208 def controller if view_context.nil? raise( ViewContextCalledBeforeRenderError, "`#controller` can't be used during initialization, as it depends " \ "on the view context that only exists once a ViewComponent is passed to " \ "the Rails render pipeline.\n\n" \ "It's sometimes possible to fix this issue by moving code dependent on " \ "`#controller` to a `#before_render` method: https://viewcomponent.org/api.html#before_render--void." ) end @__vc_controller ||= view_context.controller end |
#format ⇒ Object
For caching, such as #cache_if
265 266 267 268 |
# File 'lib/view_component/base.rb', line 265 def format # Ruby 2.6 throws a warning without checking `defined?`, 2.7 doesn't @__vc_variant if defined?(@__vc_variant) end |
#generate_distinct_locale_files ⇒ Object
Use `#generate.distinct_locale_files` instead. Will be removed in v3.0.0.
74 |
# File 'lib/view_component/base.rb', line 74 _deprecated_generate_mattr_accessor :distinct_locale_files |
#generate_locale ⇒ Object
Use `#generate.locale` instead. Will be removed in v3.0.0.
75 |
# File 'lib/view_component/base.rb', line 75 _deprecated_generate_mattr_accessor :locale |
#generate_sidecar ⇒ Object
Use `#generate.sidecar` instead. Will be removed in v3.0.0.
76 |
# File 'lib/view_component/base.rb', line 76 _deprecated_generate_mattr_accessor :sidecar |
#generate_stimulus_controller ⇒ Object
Use `#generate.stimulus_controller` instead. Will be removed in v3.0.0.
77 |
# File 'lib/view_component/base.rb', line 77 _deprecated_generate_mattr_accessor :stimulus_controller |
#helpers ⇒ ActionView::Base
A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 |
# File 'lib/view_component/base.rb', line 227 def helpers if view_context.nil? raise( ViewContextCalledBeforeRenderError, "`#helpers` can't be used during initialization, as it depends " \ "on the view context that only exists once a ViewComponent is passed to " \ "the Rails render pipeline.\n\n" \ "It's sometimes possible to fix this issue by moving code dependent on " \ "`#helpers` to a `#before_render` method: https://viewcomponent.org/api.html#before_render--void." ) end # Attempt to re-use the original view_context passed to the first # component rendered in the rendering pipeline. This prevents the # instantiation of a new view_context via `controller.view_context` which # always returns a new instance of the view context class. # # This allows ivars to remain persisted when using the same helper via # `helpers` across multiple components and partials. @__vc_helpers ||= __vc_original_view_context || controller.view_context end |
#render(options = {}, args = {}, &block) ⇒ Object
Re-use original view_context if we're not rendering a component.
This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent `view_context`,
and not the component's `view_context`.
195 196 197 198 199 200 201 202 |
# File 'lib/view_component/base.rb', line 195 def render( = {}, args = {}, &block) if .respond_to?(:set_original_view_context) .set_original_view_context(self.__vc_original_view_context) super else __vc_original_view_context.render(, args, &block) end end |
#render? ⇒ Boolean
Override to determine whether the ViewComponent should render.
180 181 182 |
# File 'lib/view_component/base.rb', line 180 def render? true end |
#render_in(view_context, &block) ⇒ String
Entrypoint for rendering components.
-
`view_context`: ActionView context from calling view
-
`block`: optional block to be captured within the view context
Returns HTML that has been escaped by the respective template handler.
87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 |
# File 'lib/view_component/base.rb', line 87 def render_in(view_context, &block) self.class.compile(raise_errors: true) @view_context = view_context self.__vc_original_view_context ||= view_context @output_buffer = ActionView::OutputBuffer.new @lookup_context ||= view_context.lookup_context # required for path helpers in older Rails versions @view_renderer ||= view_context.view_renderer # For content_for @view_flow ||= view_context.view_flow # For i18n @virtual_path ||= virtual_path # For template variants (+phone, +desktop, etc.) @__vc_variant ||= @lookup_context.variants.first # For caching, such as #cache_if @current_template = nil unless defined?(@current_template) old_current_template = @current_template @current_template = self if block && defined?(@__vc_content_set_by_with_content) raise ArgumentError, "It looks like a block was provided after calling `with_content` on #{self.class.name}, " \ "which means that ViewComponent doesn't know which content to use.\n\n" \ "To fix this issue, use either `with_content` or a block." end @__vc_content_evaluated = false @__vc_render_in_block = block before_render if render? # Ensure `content` is evaluated before rendering the template, this is # needed so slots and other side-effects are performed before the # component template is evaluated. content if self.class.use_consistent_rendering_lifecycle render_template_for(@__vc_variant).to_s + _output_postamble else "" end ensure @current_template = old_current_template end |
#render_parent ⇒ Object
Subclass components that call `super` inside their template code will cause a double render if they emit the result:
“`erb <%= super %> # double-renders <% super %> # does not double-render “`
Calls `super`, returning `nil` to avoid rendering the result twice.
148 149 150 151 152 |
# File 'lib/view_component/base.rb', line 148 def render_parent mtd = @__vc_variant ? "call_#{@__vc_variant}" : "call" method(mtd).super_method.call nil end |
#request ⇒ ActionDispatch::Request
The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.
286 287 288 |
# File 'lib/view_component/base.rb', line 286 def request @request ||= controller.request if controller.respond_to?(:request) end |
#set_original_view_context(view_context) ⇒ void
This method returns an undefined value.
Components render in their own view context. Helpers and other functionality require a reference to the original Rails view context, an instance of `ActionView::Base`. Use this method to set a reference to the original view context. Objects that implement this method will render in the component's view context, while objects that don't will render in the original view context so helpers, etc work as expected.
58 59 60 |
# File 'lib/view_component/base.rb', line 58 def set_original_view_context(view_context) self.__vc_original_view_context = view_context end |
#view_cache_dependencies ⇒ Object
For caching, such as #cache_if
258 259 260 |
# File 'lib/view_component/base.rb', line 258 def view_cache_dependencies [] end |
#virtual_path ⇒ Object
Exposes .virtual_path as an instance method
252 253 254 |
# File 'lib/view_component/base.rb', line 252 def virtual_path self.class.virtual_path end |
#with_variant(variant) ⇒ self
Will be removed in v3.0.0.
Use the provided variant instead of the one determined by the current request.
275 276 277 278 279 |
# File 'lib/view_component/base.rb', line 275 def with_variant(variant) @__vc_variant = variant self end |