Class: ViewComponent::Base

Inherits:
ActionView::Base
  • Object
show all
Includes:
ActiveSupport::Configurable, ContentAreas, Previewable, SlotableV2, Translatable, WithContentHelper
Defined in:
lib/view_component/base.rb

Constant Summary collapse

ViewContextCalledBeforeRenderError =
Class.new(StandardError)
RESERVED_PARAMETER =
:content

Constants included from Translatable

Translatable::HTML_SAFE_TRANSLATION_KEY

Constants included from SlotableV2

SlotableV2::RESERVED_NAMES

Class Attribute Summary collapse

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Methods included from WithContentHelper

#with_content

Methods included from Translatable

#html_safe_translation, #i18n_scope, #translate

Methods included from SlotableV2

#get_slot, #set_slot

Methods included from ContentAreas

#with

Constructor Details

#initializeBase

Returns a new instance of Base.



176
# File 'lib/view_component/base.rb', line 176

def initialize(*); end

Class Attribute Details

.source_locationObject



381
382
383
# File 'lib/view_component/base.rb', line 381

def source_location
  @source_location
end

.virtual_pathObject



381
382
383
# File 'lib/view_component/base.rb', line 381

def virtual_path
  @virtual_path
end

Instance Attribute Details

#__vc_original_view_contextObject

Returns the value of attribute __vc_original_view_context.



34
35
36
# File 'lib/view_component/base.rb', line 34

def __vc_original_view_context
  @__vc_original_view_context
end

Class Method Details

._after_compileObject

EXPERIMENTAL: This API is experimental and may be removed at any time. Hook for allowing components to do work as part of the compilation process.

For example, one might compile component-specific assets at this point.



41
42
43
# File 'lib/view_component/base.rb', line 41

def self._after_compile
  # noop
end

._deprecated_generate_mattr_accessor(name) ⇒ Object



48
49
50
51
52
53
54
55
# File 'lib/view_component/base.rb', line 48

def self._deprecated_generate_mattr_accessor(name)
  define_singleton_method("generate_#{name}".to_sym) do
    generate.public_send(name)
  end
  define_singleton_method("generate_#{name}=".to_sym) do |value|
    generate.public_send("#{name}=".to_sym, value)
  end
end

._sidecar_files(extensions) ⇒ Object

EXPERIMENTAL: This API is experimental and may be removed at any time. Find sidecar files for the given extensions.

The provided array of extensions is expected to contain strings starting without the “dot”, example: `[“erb”, “haml”]`.

For example, one might collect sidecar CSS files that need to be compiled.



391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
408
409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
# File 'lib/view_component/base.rb', line 391

def _sidecar_files(extensions)
  return [] unless source_location

  extensions = extensions.join(",")

  # view files in a directory named like the component
  directory = File.dirname(source_location)
  filename = File.basename(source_location, ".rb")
  component_name = name.demodulize.underscore

  # Add support for nested components defined in the same file.
  #
  # for example
  #
  # class MyComponent < ViewComponent::Base
  #   class MyOtherComponent < ViewComponent::Base
  #   end
  # end
  #
  # Without this, `MyOtherComponent` will not look for `my_component/my_other_component.html.erb`
  nested_component_files =
    if name.include?("::") && component_name != filename
      Dir["#{directory}/#{filename}/#{component_name}.*{#{extensions}}"]
    else
      []
    end

  # view files in the same directory as the component
  sidecar_files = Dir["#{directory}/#{component_name}.*{#{extensions}}"]

  sidecar_directory_files = Dir["#{directory}/#{component_name}/#{filename}.*{#{extensions}}"]

  (sidecar_files - [source_location] + sidecar_directory_files + nested_component_files).uniq
end

.collection_counter_parameterObject



586
587
588
# File 'lib/view_component/base.rb', line 586

def collection_counter_parameter
  "#{collection_parameter}_counter".to_sym
end

.collection_iteration_parameterObject



596
597
598
# File 'lib/view_component/base.rb', line 596

def collection_iteration_parameter
  "#{collection_parameter}_iteration".to_sym
end

.collection_parameterObject



577
578
579
580
581
582
583
# File 'lib/view_component/base.rb', line 577

def collection_parameter
  if provided_collection_parameter
    provided_collection_parameter
  else
    name && name.demodulize.underscore.chomp("_component").to_sym
  end
end

.compile(raise_errors: false, force: false) ⇒ Object

Compile templates to instance methods, assuming they haven't been compiled already.

Do as much work as possible in this step, as doing so reduces the amount of work done each time a component is rendered.



497
498
499
# File 'lib/view_component/base.rb', line 497

def compile(raise_errors: false, force: false)
  compiler.compile(raise_errors: raise_errors, force: force)
end

.compiled?Boolean

Returns:

  • (Boolean)


488
489
490
# File 'lib/view_component/base.rb', line 488

def compiled?
  compiler.compiled?
end

.compilerObject



502
503
504
# File 'lib/view_component/base.rb', line 502

def compiler
  @__vc_compiler ||= Compiler.new(self)
end

.counter_argument_present?Boolean

Returns:

  • (Boolean)


591
592
593
# File 'lib/view_component/base.rb', line 591

def counter_argument_present?
  initialize_parameter_names.include?(collection_counter_parameter)
end

.formatObject



513
514
515
# File 'lib/view_component/base.rb', line 513

def format
  :html
end

.identifierObject



518
519
520
# File 'lib/view_component/base.rb', line 518

def identifier
  source_location
end

.inherited(child) ⇒ Object



444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
# File 'lib/view_component/base.rb', line 444

def inherited(child)
  # Compile so child will inherit compiled `call_*` template methods that
  # `compile` defines
  compile

  # Give the child its own personal #render_template_for to protect against the case when
  # eager loading is disabled and the parent component is rendered before the child. In
  # such a scenario, the parent will override ViewComponent::Base#render_template_for,
  # meaning it will not be called for any children and thus not compile their templates.
  if !child.instance_methods(false).include?(:render_template_for) && !child.compiled?
    child.class_eval <<~RUBY, __FILE__, __LINE__ + 1
      def render_template_for(variant = nil)
        # Force compilation here so the compiler always redefines render_template_for.
        # This is mostly a safeguard to prevent infinite recursion.
        self.class.compile(raise_errors: true, force: true)
        # .compile replaces this method; call the new one
        render_template_for(variant)
      end
    RUBY
  end

  # If Rails application is loaded, add application url_helpers to the component context
  # we need to check this to use this gem as a dependency
  if defined?(Rails) && Rails.application
    child.include Rails.application.routes.url_helpers unless child < Rails.application.routes.url_helpers
  end

  # Derive the source location of the component Ruby file from the call stack.
  # We need to ignore `inherited` frames here as they indicate that `inherited`
  # has been re-defined by the consuming application, likely in ApplicationComponent.
  child.source_location = caller_locations(1, 10).reject { |l| l.label == "inherited" }[0].path

  # Removes the first part of the path and the extension.
  child.virtual_path = child.source_location.gsub(
    %r{(.*#{Regexp.quote(ViewComponent::Base.view_component_path)})|(\.rb)}, ""
  )

  # Set collection parameter to the extended component
  child.with_collection_parameter provided_collection_parameter

  super
end

.iteration_argument_present?Boolean

Returns:

  • (Boolean)


601
602
603
# File 'lib/view_component/base.rb', line 601

def iteration_argument_present?
  initialize_parameter_names.include?(collection_iteration_parameter)
end

.short_identifierObject

Provide identifier for ActionView template annotations



439
440
441
# File 'lib/view_component/base.rb', line 439

def short_identifier
  @short_identifier ||= defined?(Rails.root) ? source_location.sub("#{Rails.root}/", "") : source_location
end

.typeObject

we'll eventually want to update this to support other types



508
509
510
# File 'lib/view_component/base.rb', line 508

def type
  "text/html"
end

.validate_collection_parameter!(validate_default: false) ⇒ Object

Ensure the component initializer accepts the collection parameter. By default, we don't validate that the default parameter name is accepted, as support for collection rendering is optional.

Raises:

  • (ArgumentError)


537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
# File 'lib/view_component/base.rb', line 537

def validate_collection_parameter!(validate_default: false)
  parameter = validate_default ? collection_parameter : provided_collection_parameter

  return unless parameter
  return if initialize_parameter_names.include?(parameter)

  # If Ruby can't parse the component class, then the initalize
  # parameters will be empty and ViewComponent will not be able to render
  # the component.
  if initialize_parameters.empty?
    raise ArgumentError.new(
      "The #{self} initializer is empty or invalid." \
      "It must accept the parameter `#{parameter}` to render it as a collection.\n\n" \
      "To fix this issue, update the initializer to accept `#{parameter}`.\n\n" \
      "See https://viewcomponent.org/guide/collections.html for more information on rendering collections."
    )
  end

  raise ArgumentError.new(
    "The initializer for #{self} doesn't accept the parameter `#{parameter}`, " \
    "which is required in order to render it as a collection.\n\n" \
    "To fix this issue, update the initializer to accept `#{parameter}`.\n\n" \
    "See https://viewcomponent.org/guide/collections.html for more information on rendering collections."
  )
end

.validate_initialization_parameters!Object

Ensure the component initializer doesn't define invalid parameters that could override the framework's methods.



567
568
569
570
571
572
573
574
# File 'lib/view_component/base.rb', line 567

def validate_initialization_parameters!
  return unless initialize_parameter_names.include?(RESERVED_PARAMETER)

  raise ViewComponent::ComponentError.new(
    "#{self} initializer can't accept the parameter `#{RESERVED_PARAMETER}`, as it will override a " \
    "public ViewComponent method. To fix this issue, rename the parameter."
  )
end

.with_collection(collection, **args) ⇒ Object

Render a component for each element in a collection ([documentation](/guide/collections)):

render(ProductsComponent.with_collection(@products, foo: :bar))

Parameters:

  • collection (Enumerable)

    A list of items to pass the ViewComponent one at a time.

  • args (Arguments)

    Arguments to pass to the ViewComponent every time.



432
433
434
# File 'lib/view_component/base.rb', line 432

def with_collection(collection, **args)
  Collection.new(self, collection, **args)
end

.with_collection_parameter(parameter) ⇒ Object

Set the parameter name used when rendering elements of a collection ([documentation](/guide/collections)):

with_collection_parameter :item

Parameters:

  • parameter (Symbol)

    The parameter name used when rendering elements of a collection.



527
528
529
# File 'lib/view_component/base.rb', line 527

def with_collection_parameter(parameter)
  @provided_collection_parameter = parameter
end

Instance Method Details

#_output_postambleString

EXPERIMENTAL: Optional content to be returned after the rendered template.

Returns:

  • (String)


148
149
150
# File 'lib/view_component/base.rb', line 148

def _output_postamble
  ""
end

#before_rendervoid

This method returns an undefined value.

Called before rendering the component. Override to perform operations that depend on having access to the view context, such as helpers.



156
157
158
# File 'lib/view_component/base.rb', line 156

def before_render
  before_render_check
end

#before_render_checkvoid

Deprecated.

Use `#before_render` instead. Will be removed in v3.0.0.

This method returns an undefined value.

Called after rendering the component.



164
165
166
# File 'lib/view_component/base.rb', line 164

def before_render_check
  # noop
end

#controllerActionController::Base

The current controller. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionController::Base)


201
202
203
204
205
206
207
208
209
210
211
212
213
214
# File 'lib/view_component/base.rb', line 201

def controller
  if view_context.nil?
    raise(
      ViewContextCalledBeforeRenderError,
      "`#controller` can't be used during initialization, as it depends " \
      "on the view context that only exists once a ViewComponent is passed to " \
      "the Rails render pipeline.\n\n" \
      "It's sometimes possible to fix this issue by moving code dependent on " \
      "`#controller` to a `#before_render` method: https://viewcomponent.org/api.html#before_render--void."
    )
  end

  @__vc_controller ||= view_context.controller
end

#formatObject

For caching, such as #cache_if



258
259
260
261
262
263
# File 'lib/view_component/base.rb', line 258

def format
  # Ruby 2.6 throws a warning without checking `defined?`, 2.7 doesn't
  if defined?(@__vc_variant)
    @__vc_variant
  end
end

#generate_distinct_locale_filesObject

Deprecated.

Use `#generate.distinct_locale_files` instead. Will be removed in v3.0.0.



57
# File 'lib/view_component/base.rb', line 57

_deprecated_generate_mattr_accessor :distinct_locale_files

#generate_localeObject

Deprecated.

Use `#generate.locale` instead. Will be removed in v3.0.0.



58
# File 'lib/view_component/base.rb', line 58

_deprecated_generate_mattr_accessor :locale

#generate_sidecarObject

Deprecated.

Use `#generate.sidecar` instead. Will be removed in v3.0.0.



59
# File 'lib/view_component/base.rb', line 59

_deprecated_generate_mattr_accessor :sidecar

#generate_stimulus_controllerObject

Deprecated.

Use `#generate.stimulus_controller` instead. Will be removed in v3.0.0.



60
# File 'lib/view_component/base.rb', line 60

_deprecated_generate_mattr_accessor :stimulus_controller

#helpersActionView::Base

A proxy through which to access helpers. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionView::Base)


220
221
222
223
224
225
226
227
228
229
230
231
232
233
234
235
236
237
238
239
240
# File 'lib/view_component/base.rb', line 220

def helpers
  if view_context.nil?
    raise(
      ViewContextCalledBeforeRenderError,
      "`#helpers` can't be used during initialization, as it depends " \
      "on the view context that only exists once a ViewComponent is passed to " \
      "the Rails render pipeline.\n\n" \
      "It's sometimes possible to fix this issue by moving code dependent on " \
      "`#helpers` to a `#before_render` method: https://viewcomponent.org/api.html#before_render--void."
    )
  end

  # Attempt to re-use the original view_context passed to the first
  # component rendered in the rendering pipeline. This prevents the
  # instantiation of a new view_context via `controller.view_context` which
  # always returns a new instance of the view context class.
  #
  # This allows ivars to remain persisted when using the same helper via
  # `helpers` across multiple components and partials.
  @__vc_helpers ||= __vc_original_view_context || controller.view_context
end

#perform_renderObject



117
118
119
# File 'lib/view_component/base.rb', line 117

def perform_render
  render_template_for(@__vc_variant).to_s + _output_postamble
end

#render(options = {}, args = {}, &block) ⇒ Object

Re-use original view_context if we're not rendering a component.

This prevents an exception when rendering a partial inside of a component that has also been rendered outside of the component. This is due to the partials compiled template method existing in the parent `view_context`,

and not the component's `view_context`.


185
186
187
188
189
190
191
192
193
194
195
# File 'lib/view_component/base.rb', line 185

def render(options = {}, args = {}, &block)
  if options.respond_to?(:render_in)
    if options.is_a?(ViewComponent::Base)
      options.__vc_original_view_context = __vc_original_view_context
    end

    super
  else
    __vc_original_view_context.render(options, args, &block)
  end
end

#render?Boolean

Override to determine whether the ViewComponent should render.

Returns:

  • (Boolean)


171
172
173
# File 'lib/view_component/base.rb', line 171

def render?
  true
end

#render_in(view_context, &block) ⇒ String

Entrypoint for rendering components.

  • `view_context`: ActionView context from calling view

  • `block`: optional block to be captured within the view context

Returns HTML that has been escaped by the respective template handler.

Returns:

  • (String)


70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
# File 'lib/view_component/base.rb', line 70

def render_in(view_context, &block)
  @view_context = view_context
  self.__vc_original_view_context ||= view_context

  @output_buffer = ActionView::OutputBuffer.new unless defined?(@global_buffer_in_use) && @global_buffer_in_use

  @lookup_context ||= view_context.lookup_context

  # required for path helpers in older Rails versions
  @view_renderer ||= view_context.view_renderer

  # For content_for
  @view_flow ||= view_context.view_flow

  # For i18n
  @virtual_path ||= virtual_path

  # For template variants (+phone, +desktop, etc.)
  @__vc_variant ||= @lookup_context.variants.first

  # For caching, such as #cache_if
  @current_template = nil unless defined?(@current_template)
  old_current_template = @current_template
  @current_template = self

  if block && defined?(@__vc_content_set_by_with_content)
    raise ArgumentError.new(
      "It looks like a block was provided after calling `with_content` on #{self.class.name}, " \
      "which means that ViewComponent doesn't know which content to use.\n\n" \
      "To fix this issue, use either `with_content` or a block."
    )
  end

  @__vc_content_evaluated = false
  @__vc_render_in_block = block

  before_render

  if render?
    perform_render
  else
    ""
  end
ensure
  @current_template = old_current_template
end

#render_parentObject

Subclass components that call `super` inside their template code will cause a double render if they accidentally emit the result:

<%= super %> # double-renders

<% super %> # does not double-render

Calls `super`, returning nil to avoid rendering the result twice.



129
130
131
132
133
# File 'lib/view_component/base.rb', line 129

def render_parent
  mtd = @__vc_variant ? "call_#{@__vc_variant}" : "call"
  method(mtd).super_method.call
  nil
end

#render_template_for(variant = nil) ⇒ Object

:nocov:



136
137
138
139
140
141
142
# File 'lib/view_component/base.rb', line 136

def render_template_for(variant = nil)
  # Force compilation here so the compiler always redefines render_template_for.
  # This is mostly a safeguard to prevent infinite recursion.
  self.class.compile(raise_errors: true, force: true)
  # .compile replaces this method; call the new one
  render_template_for(variant)
end

#requestActionDispatch::Request

The current request. Use sparingly as doing so introduces coupling that inhibits encapsulation & reuse, often making testing difficult.

Returns:

  • (ActionDispatch::Request)


281
282
283
# File 'lib/view_component/base.rb', line 281

def request
  @request ||= controller.request if controller.respond_to?(:request)
end

#view_cache_dependenciesObject

For caching, such as #cache_if



251
252
253
# File 'lib/view_component/base.rb', line 251

def view_cache_dependencies
  []
end

#virtual_pathObject

Exposes .virtual_path as an instance method



245
246
247
# File 'lib/view_component/base.rb', line 245

def virtual_path
  self.class.virtual_path
end

#with_variant(variant) ⇒ self

Deprecated.

Will be removed in v3.0.0.

Use the provided variant instead of the one determined by the current request.

Parameters:

  • variant (Symbol)

    The variant to be used by the component.

Returns:

  • (self)


270
271
272
273
274
# File 'lib/view_component/base.rb', line 270

def with_variant(variant)
  @__vc_variant = variant

  self
end