Module: ViewComponent::TestHelpers

Included in:
TestCase
Defined in:
lib/view_component/test_helpers.rb

Instance Attribute Summary collapse

Instance Method Summary collapse

Instance Attribute Details

#rendered_contentObject (readonly)



32
33
34
# File 'lib/view_component/test_helpers.rb', line 32

def rendered_content
  @rendered_content
end

Instance Method Details

#build_controller(klass) ⇒ Object



208
209
210
# File 'lib/view_component/test_helpers.rb', line 208

def build_controller(klass)
  klass.new.tap { |c| c.request = request }.extend(Rails.application.routes.url_helpers)
end

#controllerObject



126
127
128
# File 'lib/view_component/test_helpers.rb', line 126

def controller
  @controller ||= build_controller(Base.test_controller.constantize)
end

#render_in_view_context(&block) ⇒ Object

Execute the given block in the view context. Internally sets `page` to be a `Capybara::Node::Simple`, allowing for Capybara assertions to be used:

“`ruby render_in_view_context do

render(MyComponent.new)

end

assert_text(“Hello, World!”) “`



119
120
121
122
123
# File 'lib/view_component/test_helpers.rb', line 119

def render_in_view_context(&block)
  @page = nil
  @rendered_content = controller.view_context.instance_exec(&block)
  Nokogiri::HTML.fragment(@rendered_content)
end

#render_inline(component, **args, &block) ⇒ Nokogiri::HTML

Render a component inline. Internally sets `page` to be a `Capybara::Node::Simple`, allowing for Capybara assertions to be used:

“`ruby render_inline(MyComponent.new) assert_text(“Hello, World!”) “`

Parameters:

Returns:

  • (Nokogiri::HTML)


56
57
58
59
60
61
62
63
64
65
66
# File 'lib/view_component/test_helpers.rb', line 56

def render_inline(component, **args, &block)
  @page = nil
  @rendered_content =
    if Rails.version.to_f >= 6.1
      controller.view_context.render(component, args, &block)
    else
      controller.view_context.render_component(component, &block)
    end

  Nokogiri::HTML.fragment(@rendered_content)
end

#render_preview(name) ⇒ Nokogiri::HTML

Render a preview inline. Internally sets `page` to be a `Capybara::Node::Simple`, allowing for Capybara assertions to be used:

“`ruby render_preview(:default) assert_text(“Hello, World!”) “`

Note: `#rendered_preview` expects a preview to be defined with the same class name as the calling test, but with `Test` replaced with `Preview`:

MyComponentTest -> MyComponentPreview etc.

In RSpec, `Preview` is appended to `described_class`.

Parameters:

  • preview (String)

    The name of the preview to be rendered.

Returns:

  • (Nokogiri::HTML)


85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
# File 'lib/view_component/test_helpers.rb', line 85

def render_preview(name)
  begin
    preview_klass = if respond_to?(:described_class)
      raise "`render_preview` expected a described_class, but it is nil." if described_class.nil?

      "#{described_class}Preview"
    else
      self.class.name.gsub("Test", "Preview")
    end
    preview_klass = preview_klass.constantize
  rescue NameError
    raise NameError, "`render_preview` expected to find #{preview_klass}, but it does not exist."
  end

  previews_controller = build_controller(Rails.application.config.view_component.preview_controller.constantize)
  previews_controller.request.params[:path] = "#{preview_klass.preview_name}/#{name}"
  previews_controller.response = ActionDispatch::Response.new
  result = previews_controller.previews

  @rendered_content = result

  Nokogiri::HTML.fragment(@rendered_content)
end

#rendered_componentString

Returns the result of a render_inline call.

Returns:

  • (String)


37
38
39
40
41
42
43
44
# File 'lib/view_component/test_helpers.rb', line 37

def rendered_component
  ViewComponent::Deprecation.warn(
    "`rendered_component` is deprecated and will be removed in v3.0.0. " \
    "Use `page` instead."
  )

  rendered_content
end

#requestObject



131
132
133
134
135
136
137
138
# File 'lib/view_component/test_helpers.rb', line 131

def request
  @request ||=
    begin
      request = ActionDispatch::TestRequest.create
      request.session = ActionController::TestSession.new
      request
    end
end

#with_controller_class(klass) ⇒ Object

Set the controller to be used while executing the given block, allowing access to controller-specific methods:

“`ruby with_controller_class(UsersController) do

render_inline(MyComponent.new)

end “`

Parameters:

  • klass (ActionController::Base)

    The controller to be used.



168
169
170
171
172
173
174
175
# File 'lib/view_component/test_helpers.rb', line 168

def with_controller_class(klass)
  old_controller = defined?(@controller) && @controller

  @controller = build_controller(klass)
  yield
ensure
  @controller = old_controller
end

#with_request_url(path) ⇒ Object

Set the URL of the current request (such as when using request-dependent path helpers):

“`ruby with_request_url(“/users/42”) do

render_inline(MyComponent.new)

end “`

Parameters:

  • path (String)

    The path to set for the current request.



186
187
188
189
190
191
192
193
194
195
196
197
198
199
200
201
202
203
204
205
# File 'lib/view_component/test_helpers.rb', line 186

def with_request_url(path)
  old_request_path_info = request.path_info
  old_request_path_parameters = request.path_parameters
  old_request_query_parameters = request.query_parameters
  old_request_query_string = request.query_string
  old_controller = defined?(@controller) && @controller

  path, query = path.split("?", 2)
  request.path_info = path
  request.path_parameters = Rails.application.routes.recognize_path(path)
  request.set_header("action_dispatch.request.query_parameters", Rack::Utils.parse_nested_query(query))
  request.set_header(Rack::QUERY_STRING, query)
  yield
ensure
  request.path_info = old_request_path_info
  request.path_parameters = old_request_path_parameters
  request.set_header("action_dispatch.request.query_parameters", old_request_query_parameters)
  request.set_header(Rack::QUERY_STRING, old_request_query_string)
  @controller = old_controller
end

#with_variant(variant) ⇒ Object

Set the Action Pack request variant for the given block:

“`ruby with_variant(:phone) do

render_inline(MyComponent.new)

end “`

Parameters:

  • variant (Symbol)

    The variant to be set for the provided block.



149
150
151
152
153
154
155
156
# File 'lib/view_component/test_helpers.rb', line 149

def with_variant(variant)
  old_variants = controller.view_context.lookup_context.variants

  controller.view_context.lookup_context.variants = variant
  yield
ensure
  controller.view_context.lookup_context.variants = old_variants
end